import glob
import os
import pathlib
from xml.etree import ElementTree as ET

import numpy as np
from dm_control import mjcf
from typing_extensions import Union, List, Optional, Dict, Tuple, Self

from ..datastructures.dataclasses import Color, VisualShape, BoxVisualShape, CylinderVisualShape, \
    SphereVisualShape, MeshVisualShape, VisualShapeUnion
from ..datastructures.enums import JointType, MJCFGeomType, MJCFJointType, Shape
from ..datastructures.pose import PoseStamped, Point
from ..description import JointDescription as AbstractJointDescription, \
    LinkDescription as AbstractLinkDescription, ObjectDescription as AbstractObjectDescription, ObjectDescription
from ..failures import MultiplePossibleTipLinks
from ..ros import  get_parameter

try:
    from multiverse_parser import Configuration, Factory, InertiaSource, GeomBuilder
    from multiverse_parser import (WorldBuilder,
                                   GeomType, GeomProperty,
                                   MeshProperty,
                                   MaterialProperty)
    from multiverse_parser import MjcfExporter
    from pxr import Usd, UsdGeom
except ImportError:
    # do not import this module if multiverse is not found
    raise ImportError("Multiverse not found.")


class LinkDescription(AbstractLinkDescription):
    """
    A class that represents a link description of an object.
    """

    def __init__(self, mjcf_description: mjcf.Element, mesh_dir: Optional[str] = None):
        super().__init__(mjcf_description)
        self.mesh_dir = mesh_dir

    @property
    def geometry(self) -> Union[List[VisualShape], VisualShape, None]:
        """
        :return: The geometry type of the collision element of this link.
        """
        all_geoms = self.parsed_description.find_all('geom')
        if len(all_geoms) == 0:
            return None
        elif len(all_geoms) == 1:
            return self._get_visual_shape(all_geoms[0])
        else:
            return [self._get_visual_shape(geom) for geom in all_geoms]

    def _get_visual_shape(self, mjcf_geometry) -> Union[VisualShape, None]:
        """
        :param mjcf_geometry: The MJCFGeometry to get the visual shape for.
        :return: The VisualShape of the given MJCFGeometry object.
        """
        if mjcf_geometry.type == MJCFGeomType.BOX.value:
            return BoxVisualShape(Color(), [0, 0, 0], mjcf_geometry.size)
        if mjcf_geometry.type == MJCFGeomType.CYLINDER.value:
            return CylinderVisualShape(Color(), [0, 0, 0], mjcf_geometry.size[0], mjcf_geometry.size[1] * 2)
        if mjcf_geometry.type == MJCFGeomType.SPHERE.value:
            return SphereVisualShape(Color(), [0, 0, 0], mjcf_geometry.size[0])
        if mjcf_geometry.type == MJCFGeomType.MESH.value:
            mesh_filename = mjcf_geometry.mesh.file.prefix + mjcf_geometry.mesh.file.extension
            mesh_filename = self.look_for_file_in_mesh_dir(mesh_filename)
            return MeshVisualShape(Color(), [0, 0, 0], mjcf_geometry.mesh.scale, mesh_filename)
        return None

    def look_for_file_in_mesh_dir(self, file_name: str) -> str:
        """
        Look for a file in the mesh directory of the object.

        :param file_name: The name of the file.
        :return: The path to the file.
        """
        # search for the file in the mesh directory
        if self.mesh_dir is not None:
            if os.path.exists(os.path.join(self.mesh_dir, file_name)):
                return os.path.join(self.mesh_dir, file_name)
            else:
                # use glob to search for the file in the mesh directory
                files = glob.glob(os.path.join(self.mesh_dir, '**', file_name), recursive=True)
                if len(files) > 0:
                    return files[0]

    @property
    def origin(self) -> Union[PoseStamped, None]:
        """
        :return: The origin of this link.
        """
        return parse_pose_from_body_element(self.parsed_description)

    @property
    def name(self) -> str:
        return self.parsed_description.name


class JointDescription(AbstractJointDescription):
    mjcf_type_map = {
        MJCFJointType.HINGE.value: JointType.REVOLUTE,
        MJCFJointType.BALL.value: JointType.SPHERICAL,
        MJCFJointType.SLIDE.value: JointType.PRISMATIC,
        MJCFJointType.FREE.value: JointType.FLOATING
    }
    """
    A dictionary mapping the MJCF joint types to the PyCRAM joint types.
    """

    pycram_type_map = {pycram_type: mjcf_type for mjcf_type, pycram_type in mjcf_type_map.items()}
    """
    A dictionary mapping the PyCRAM joint types to the MJCF joint types.
    """

    def __init__(self, mjcf_description: mjcf.Element, is_virtual: Optional[bool] = False):
        super().__init__(mjcf_description, is_virtual=is_virtual)

    @property
    def origin(self) -> PoseStamped:
        return parse_pose_from_body_element(self.parsed_description)

    @property
    def name(self) -> str:
        return self.parsed_description.name

    @property
    def has_limits(self) -> bool:
        return self.parsed_description.limited

    @property
    def type(self) -> JointType:
        """
        :return: The type of this joint.
        """
        if hasattr(self.parsed_description, 'type'):
            return self.mjcf_type_map[self.parsed_description.type]
        else:
            return self.mjcf_type_map[MJCFJointType.FREE.value]

    @property
    def axis(self) -> Point:
        """
        :return: The axis of this joint, for example the rotation axis for a revolute joint.
        """
        return Point(**dict(zip(["x", "y", "z"], self.parsed_description)))

    @property
    def lower_limit(self) -> Union[float, None]:
        """
        :return: The lower limit of this joint, or None if the joint has no limits.
        """
        if self.has_limits:
            return self.parsed_description.range[0]
        else:
            return None

    @property
    def upper_limit(self) -> Union[float, None]:
        """
        :return: The upper limit of this joint, or None if the joint has no limits.
        """
        if self.has_limits:
            return self.parsed_description.range[1]
        else:
            return None

    @property
    def parent(self) -> str:
        """
        :return: The name of the parent link of this joint.
        """
        return self._parent_link_element.parent.name

    @property
    def child(self) -> str:
        """
        :return: The name of the child link of this joint.
        """
        return self._parent_link_element.name

    @property
    def _parent_link_element(self) -> mjcf.Element:
        return self.parsed_description.parent

    @property
    def damping(self) -> float:
        """
        :return: The damping of this joint.
        """
        return self.parsed_description.damping

    @property
    def friction(self) -> float:
        raise NotImplementedError("Friction is not implemented for MJCF joints.")


class ObjectFactory(Factory):
    """
    Create MJCF object descriptions from mesh files.
    """

    def __init__(self, file_path: str, config: Configuration):
        super().__init__(file_path, config)

    def from_mesh_file(self, object_name: str, texture_type: str = "png"):

        self._world_builder = WorldBuilder(usd_file_path=self.tmp_usd_file_path)

        body_builder = self._world_builder.add_body(body_name=object_name)

        tmp_usd_mesh_file_path, tmp_origin_mesh_file_path = self.import_mesh(
            mesh_file_path=self.source_file_path, merge_mesh=True)
        mesh_stage = Usd.Stage.Open(tmp_usd_mesh_file_path)
        for idx, mesh_prim in enumerate([prim for prim in mesh_stage.Traverse() if prim.IsA(UsdGeom.Mesh)]):
            mesh_name = mesh_prim.GetName()
            mesh_path = mesh_prim.GetPath()
            mesh_property = MeshProperty.from_mesh_file_path(mesh_file_path=tmp_usd_mesh_file_path,
                                                             mesh_path=mesh_path)
            # mesh_property._texture_coordinates = None # TODO: See if needed otherwise remove it.
            geom_property = GeomProperty(geom_type=GeomType.MESH,
                                         is_visible=False,
                                         is_collidable=True)
            geom_builder = body_builder.add_geom(geom_name=f"SM_{object_name}_mesh_{idx}",
                                                 geom_property=geom_property)
            geom_builder.add_mesh(mesh_name=mesh_name, mesh_property=mesh_property)

            # Add texture if available
            texture_file_path = self.source_file_path.replace(pathlib.Path(self.source_file_path).suffix,
                                                              f".{texture_type}")
            if pathlib.Path(texture_file_path).exists():
                self.add_material_with_texture(geom_builder=geom_builder, material_name=f"M_{object_name}_{idx}",
                                              texture_file_path=texture_file_path)

            geom_builder.build()

        body_builder.compute_and_set_inertial(inertia_source=InertiaSource.FROM_COLLISION_MESH)

    @staticmethod
    def add_material_with_texture(geom_builder: GeomBuilder, material_name: str, texture_file_path: str):
        """
        Add a material with a texture to the geom builder.

        :param geom_builder: The geom builder to add the material to.
        :param material_name: The name of the material.
        :param texture_file_path: The path to the texture file.
        """
        material_property = MaterialProperty(diffuse_color=texture_file_path,
                                             opacity=None,
                                             emissive_color=None,
                                             specular_color=None)
        geom_builder.add_material(material_name=material_name,
                                  material_property=material_property)

    def export_to_mjcf(self, output_file_path: str):
        """
        Export the object to a MJCF file.

        :param output_file_path: The path to the output file.
        """
        exporter = MjcfExporter(self, output_file_path)
        exporter.build()
        exporter.export(keep_usd=False)


class PrimitiveObjectFactory(ObjectFactory):

    def __init__(self, object_name: str, shape_data: VisualShapeUnion, save_path: str,
                 orientation: Optional[List[float]] = None):
        """
        Create an MJCF object description from a primitive shape.

        :param object_name: The name of the object.
        :param shape_data: The shape data of the object.
        :param save_path: The path to save the MJCF file.
        :param orientation: The orientation of the object.
        """
        self.shape_data: VisualShapeUnion = shape_data
        self.orientation: List[float] = [0, 0, 0, 1] if orientation is None else orientation
        config = Configuration(model_name=object_name,
                               fixed_base=False,
                               with_visual=True,
                               with_collision=True,
                               default_rgba=np.array(shape_data.rgba_color.get_rgba()))
        super().__init__(save_path, config)

    def build_shape(self):

        self._world_builder = WorldBuilder(usd_file_path=self.tmp_usd_file_path)

        body_builder = self._world_builder.add_body(body_name=self.config.model_name)

        geom_type_map = {
            Shape.SPHERE: GeomType.SPHERE,
            Shape.BOX: GeomType.CUBE,
            Shape.CYLINDER: GeomType.CYLINDER,
            Shape.PLANE: GeomType.PLANE,
            Shape.CAPSULE: GeomType.CAPSULE,
        }
        geom_property = GeomProperty(geom_type=geom_type_map[self.shape_data.visual_geometry_type],
                                     is_visible=self.config.with_visual,
                                     is_collidable=self.config.with_collision,
                                     rgba=self.config.default_rgba)

        geom_builder = body_builder.add_geom(
            geom_name=f"{self.config.model_name}_Shape",
            geom_property=geom_property
        )
        geom_pos = self.shape_data.visual_frame_position
        geom_quat = np.array(self.orientation)
        if self.shape_data.visual_geometry_type == Shape.PLANE:
            geom_builder.set_transform(pos=geom_pos, quat=geom_quat, scale=np.array([50, 50, 1]))
        elif self.shape_data.visual_geometry_type == Shape.BOX:
            geom_builder.set_transform(pos=geom_pos, quat=geom_quat, scale=np.array(self.shape_data.half_extents) * 2)
        elif self.shape_data.visual_geometry_type == Shape.SPHERE:
            geom_builder.set_transform(pos=geom_pos, quat=geom_quat)
            geom_builder.set_attribute(radius=self.shape_data.radius)
        elif self.shape_data.visual_geometry_type in [Shape.CYLINDER, Shape.CAPSULE]:
            geom_builder.set_transform(pos=geom_pos, quat=geom_quat)
            geom_builder.set_attribute(radius=self.shape_data.radius, height=self.shape_data.length)
        geom_builder.build()


class ObjectDescription(AbstractObjectDescription):
    """
    A class that represents an object description of an object.
    """

    COMPILER_TAG = 'compiler'
    """
    The tag of the compiler element in the MJCF file.
    """
    MESH_DIR_ATTR = 'meshdir'
    TEXTURE_DIR_ATTR = 'texturedir'
    """
    The attributes of the compiler element in the MJCF file. The meshdir attribute is the directory where the mesh files
    are stored and the texturedir attribute is the directory where the texture files are stored."""

    class Link(AbstractObjectDescription.Link, LinkDescription):
        ...

    class RootLink(AbstractObjectDescription.RootLink, Link):
        ...

    class Joint(AbstractObjectDescription.Joint, JointDescription):
        ...

    def __init__(self):
        super().__init__()
        self._link_map = None
        self._joint_map = None
        self._child_map = None
        self._parent_map = None
        self._links = None
        self._joints = None
        self.virtual_joint_names = []
        self._meshes_dir: Optional[str] = None

    def merge_description(self, other: ObjectDescription, parent_link: Optional[str] = None,
                          child_link: Optional[str] = None,
                          joint_type: JointType = JointType.FIXED,
                          axis: Optional[Point] = None,
                          lower_limit: Optional[float] = None, upper_limit: Optional[float] = None,
                          child_pose_wrt_parent: Optional[PoseStamped] = None,
                          in_place: bool = False,
                          new_description_file: Optional[str] = None) -> Union[ObjectDescription, Self]:
        raise NotImplementedError

    @property
    def mesh_dir(self) -> Optional[str]:
        try:
            return getattr(self.parsed_description.compiler, self.MESH_DIR_ATTR)
        except AttributeError:
            return None

    @property
    def child_map(self) -> Dict[str, List[Tuple[str, str]]]:
        """
        :return: A dictionary mapping the name of a link to its children which are represented as a tuple of the child
            joint name and the link name.
        """
        if self._child_map is None:
            self._child_map = self._construct_child_map()
        return self._child_map

    def _construct_child_map(self) -> Dict[str, List[Tuple[str, str]]]:
        """
        Construct the child map of the object.
        """
        child_map = {}
        for joint in self.joints:
            if joint.parent not in child_map:
                child_map[joint.parent] = [(joint.name, joint.child)]
            else:
                child_map[joint.parent].append((joint.name, joint.child))
        return child_map

    @property
    def parent_map(self) -> Dict[str, Tuple[str, str]]:
        """
        :return: A dictionary mapping the name of a link to its parent joint and link as a tuple.
        """
        if self._parent_map is None:
            self._parent_map = self._construct_parent_map()
        return self._parent_map

    def _construct_parent_map(self) -> Dict[str, Tuple[str, str]]:
        """
        Construct the parent map of the object.
        """
        child_map = self.child_map
        parent_map = {}
        for parent, children in child_map.items():
            for child in children:
                parent_map[child[1]] = (child[0], parent)
        return parent_map

    @property
    def link_map(self) -> Dict[str, LinkDescription]:
        """
        :return: A dictionary mapping the name of a link to its description.
        """
        if self._link_map is None:
            self._link_map = {link.name: link for link in self.links}
        return self._link_map

    @property
    def joint_map(self) -> Dict[str, JointDescription]:
        """
        :return: A dictionary mapping the name of a joint to its description.
        """
        if self._joint_map is None:
            self._joint_map = {joint.name: joint for joint in self.joints}
        return self._joint_map

    def add_joint(self, name: str, child: str, joint_type: JointType,
                  axis: Point, parent: Optional[str] = None, origin: Optional[PoseStamped] = None,
                  lower_limit: Optional[float] = None, upper_limit: Optional[float] = None,
                  is_virtual: Optional[bool] = False) -> None:
        """
        Finds the child link and adds a joint to it in the object description.
        for arguments documentation see :meth:`pycram.description.ObjectDescription.add_joint`
        """

        position: Optional[List[float]] = None
        quaternion: Optional[List[float]] = None
        lower_limit: float = 0.0 if lower_limit is None else lower_limit
        upper_limit: float = 0.0 if upper_limit is None else upper_limit
        limit = [lower_limit, upper_limit]

        if origin is not None:
            position = origin.position.to_list()
            quaternion = origin.orientation.to_list()
            quaternion = [quaternion[1], quaternion[2], quaternion[3], quaternion[0]]
        if axis is not None:
            axis = [axis.x, axis.y, axis.z]
        self.parsed_description.find(child).add('joint', name=name, type=JointDescription.pycram_type_map[joint_type],
                                                axis=axis, pos=position, quat=quaternion, range=limit)
        if is_virtual:
            self.virtual_joint_names.append(name)

    def load_description(self, path) -> mjcf.RootElement:
        return mjcf.from_file(path, model_dir=pathlib.Path(path).parent)

    def load_description_from_string(self, description_string: str) -> mjcf.RootElement:
        return mjcf.from_xml_string(description_string)

    def generate_from_mesh_file(self, path: str, name: str, color: Optional[Color] = Color(),
                                save_path: Optional[str] = None) -> None:
        """
        Generate a mjcf xml file with the given .obj or .stl file as mesh. In addition, use the given rgba_color
         to create a material tag in the xml.

        :param path: The path to the mesh file.
        :param name: The name of the object.
        :param color: The color of the object.
        :param save_path: The path to save the generated xml file.
        """
        factory = ObjectFactory(file_path=path,
                                config=Configuration(model_name=name,
                                                     fixed_base=False,
                                                     default_rgba=np.array(color.get_rgba())))
        factory.from_mesh_file(object_name=name)
        factory.export_to_mjcf(output_file_path=save_path)

    def generate_from_description_file(self, path: str, save_path: str, make_mesh_paths_absolute: bool = True) -> None:
        model_str = self.replace_relative_paths_with_absolute_paths(path)
        self.write_description_to_file(model_str, save_path)

    def replace_relative_paths_with_absolute_paths(self, model_path: str) -> str:
        """
        Replace the relative paths in the xml file to be absolute paths.

        :param model_path: The path to the xml file.
        """
        tree = ET.parse(model_path)
        root = tree.getroot()
        compiler = root.find(self.COMPILER_TAG)
        model_dir = pathlib.Path(model_path).parent
        for rel_dir_attrib in [self.MESH_DIR_ATTR, self.TEXTURE_DIR_ATTR]:
            rel_dir = compiler.get(rel_dir_attrib)
            abs_dir = str(pathlib.Path(os.path.join(model_dir, rel_dir)).resolve())
            if rel_dir_attrib == self.MESH_DIR_ATTR:
                self._meshes_dir = abs_dir
            compiler.set(rel_dir_attrib, abs_dir)
        return ET.tostring(root, encoding='unicode', method='xml')

    def generate_from_parameter_server(self, name: str, save_path: str) -> None:
        mjcf_string = get_parameter(name)
        self.write_description_to_file(mjcf_string, save_path)

    @property
    def joints(self) -> List[JointDescription]:
        """
        :return: A list of joints descriptions of this object.
        """
        if self._joints is None:
            self._joints = [JointDescription(joint) for joint in self.parsed_description.find_all('joint')]
        return self._joints

    @property
    def links(self) -> List[LinkDescription]:
        """
        :return: A list of link descriptions of this object.
        """
        if self._links is None:
            self._links = [LinkDescription(link, self.mesh_dir) for link in self.parsed_description.find_all('body')]
        return self._links

    def get_root(self) -> str:
        """
        :return: the name of the root link of this object.
        """
        if len(self.links) == 1:
            return self.links[0].name
        elif len(self.links) > 1:
            return self.links[1].name
        else:
            raise ValueError("No links found in the object description.")

    def get_tip(self) -> str:
        """
        :return: the name of the tip link of this object.
        :raises MultiplePossibleTipLinks: If there are multiple possible tip links.
        """
        link = self.get_root()
        while link in self.child_map:
            children = self.child_map[link]
            if len(children) > 1:
                # Multiple children, can't decide which one to take (e.g. fingers of a hand)
                raise MultiplePossibleTipLinks(self.name, link, [child[1] for child in children])
            else:
                child = children[0][1]
                link = child
        return link

    def get_chain(self, start_link_name: str, end_link_name: str, joints: Optional[bool] = True,
                  links: Optional[bool] = True, fixed: Optional[bool] = True) -> List[str]:
        """
        :param start_link_name: The name of the start link of the chain.
        :param end_link_name: The name of the end link of the chain.
        :param joints: Whether to include joints in the chain.
        :param links: Whether to include links in the chain.
        :param fixed: Whether to include fixed joints in the chain (Note: not used in MJCF).
        :return: the chain of links from 'start_link_name' to 'end_link_name'.
        """
        chain = []
        if links:
            chain.append(end_link_name)
        link = end_link_name
        while link != start_link_name:
            (joint, parent) = self.parent_map[link]
            if joints:
                chain.append(joint)
            if links:
                chain.append(parent)
            link = parent
        chain.reverse()
        return chain

    @staticmethod
    def get_file_extension() -> str:
        """
        :return: The file extension of the URDF file.
        """
        return '.xml'

    @property
    def origin(self) -> PoseStamped:
        return parse_pose_from_body_element(self.parsed_description)

    @property
    def name(self) -> str:
        return self.parsed_description.name


def parse_pose_from_body_element(body: mjcf.Element) -> PoseStamped:
    """
    Parse the pose from a body element.

    :param body: The body element.
    :return: The pose of the body.
    """
    position = body.pos
    quaternion = body.quat
    position = [0, 0, 0] if position is None else position
    quaternion = [1, 0, 0, 0] if quaternion is None else quaternion
    quaternion = [quaternion[1], quaternion[2], quaternion[3], quaternion[0]]
    return PoseStamped(position, quaternion)



import os
from functools import lru_cache


@lru_cache(maxsize=None)
def get_ros_package_path(package_name: str) -> str:
    """
    Get the path of a ROS package. Using the os module to avoid importing rospkg.
    """
    home_dir = os.path.expanduser('~')
    # search recursively in the home directory
    for root, dirs, files in os.walk(home_dir):
        if package_name in dirs:
            return os.path.join(root, package_name)
    raise FileNotFoundError(f"Package '{package_name}' not found in the home directory.")

def sleep(duration: float):
    time.sleep(duration)

def get_node_names(namespace=None):
    """
    Get the names of all nodes in the ROS system.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to get the node names.
    return []

def create_ros_pack(ros_paths=None):
    """
    Creates a RosPack instance to search for resources of ros packages.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to create a RosPack instance.
    return None

class ResourceNotFound(Exception):
    """
    Exception raised when a resource is not found.
    """
    pass

def get_parameter(name: str):
    """
    Get a parameter from the ROS parameter server.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to get the parameter.
    return None

def wait_for_message(topic_name: str, msg_type):
    """
    Wait for a message on a topic.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to wait for the message.
    return None

def is_master_online():
    """
    Check if the ROS master is online.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to check if the master is online.
    return False


def get_time():
    """
    Get the current time from the ROS system.
    """
    # This is a placeholder implementation. In a real implementation, you would use
    # the appropriate ROS API to get the time.
    return 0.0


import rosservice
import rospy


def get_service_proxy(topic_name: str, service_message) -> rospy.ServiceProxy:
    return rospy.ServiceProxy(topic_name, service_message)


def wait_for_service(topic_name: str):
    rospy.loginfo_once(f"Waiting for service: {topic_name}")
    rospy.wait_for_service(topic_name)



import atexit
import threading
import time
from functools import cached_property
from typing import List, Optional, Tuple

import numpy as np
from geometry_msgs.msg import Vector3, Point
from std_msgs.msg import ColorRGBA
from visualization_msgs.msg import Marker, MarkerArray

from ..datastructures.dataclasses import BoxVisualShape, CylinderVisualShape, MeshVisualShape, SphereVisualShape
from ..datastructures.pose import PoseStamped, TransformStamped
from ..datastructures.world import World
from ..designator import ObjectDesignatorDescription
from ..ros import  Duration, Time
from ..ros import  loginfo, logwarn, logerr
from ..ros import  create_publisher
from ..ros import  sleep


class VizMarkerPublisher:
    """
    Publishes an Array of visualization marker which represent the situation in the World
    """

    def __init__(self, topic_name="/pycram/viz_marker", interval=0.1, reference_frame="map", use_prospection_world=False, publish_visuals=False):
        """
        The Publisher creates an Array of Visualization marker with a Marker for each link of each Object in the
        World. This Array is published with a rate of interval.

        :param topic_name: The name of the topic to which the Visualization Marker should be published.
        :param interval: The interval at which the visualization marker should be published, in seconds.
        :param reference_frame: The reference frame of the visualization marker.
        :param use_prospection_world: If True, the visualization marker will be published for the prospection world.
        :param publish_visuals: If True, the visualization marker will be published.
        """
        self.use_prospection_world = use_prospection_world
        if self.use_prospection_world:
            self.topic_name = "/pycram/prospection_viz_marker"
        else:
            self.topic_name = topic_name
        self.interval = interval
        self.reference_frame = reference_frame

        self.pub = create_publisher(self.topic_name, MarkerArray, queue_size=10)

        self.thread = threading.Thread(target=self._publish)
        self.kill_event = threading.Event()
        self.publish_visuals = publish_visuals
        if self.use_prospection_world:
            self.main_world = World.current_world.prospection_world
        else:
            self.main_world = World.current_world if not World.current_world.is_prospection_world else World.current_world.world_sync.world
        self.lock = self.main_world.object_lock
        self.thread.start()
        atexit.register(self._stop_publishing)

    def _publish(self) -> None:
        """
        Constantly publishes the Marker Array. To the given topic name at a fixed rate.
        """
        while not self.kill_event.is_set():
            self.lock.acquire()
            marker_array = self._make_marker_array()
            self.lock.release()
            self.pub.publish(marker_array)
            time.sleep(self.interval)

    def _make_marker_array(self) -> MarkerArray:
        """
        Creates the Marker Array to be published. There is one Marker for link for each object in the Array, each Object
        creates a name space in the visualization Marker. The type of Visualization Marker is decided by the collision
        tag of the URDF.

        :return: An Array of Visualization Marker
        """
        marker_array = MarkerArray()
        for obj in self.main_world.objects:
            if obj.name == "floor":
                continue
            for link in obj.link_name_to_id.keys():
                if self.publish_visuals:
                    geoms = obj.get_link_visual_geometry(link)
                else:
                    geoms = obj.get_link_geometry(link)
                if not isinstance(geoms, list):
                    geoms = [geoms]
                geom = geoms[0] if len(geoms) > 0 else None
                if not geom:
                    continue
                msg = Marker()
                msg.header.frame_id = self.reference_frame
                msg.ns = obj.name
                msg.id = obj.link_name_to_id[link]
                msg.type = Marker.MESH_RESOURCE
                msg.action = Marker.ADD
                link_pose = obj.get_link_transform(link)
                if obj.get_link_origin(link) is not None:
                    link_origin = obj.get_link_origin_transform(link)
                else:
                    link_origin = TransformStamped.from_list()
                link_pose_with_origin = link_pose * link_origin
                msg.pose = link_pose_with_origin.to_pose_stamped().pose.ros_message()

                color = obj.get_link_color(link).get_rgba()

                msg.color = ColorRGBA(**dict(zip(["r", "g", "b","a"], color)))
                if self.use_prospection_world:
                    msg.color.a = 0.5
                msg.lifetime = Duration(1)

                if isinstance(geom, MeshVisualShape):
                    msg.type = Marker.MESH_RESOURCE
                    msg.mesh_resource = "file://" + geom.file_name
                    if hasattr(geom, "scale") and geom.scale is not None:
                        msg.scale = Vector3(**dict(zip(["x", "y", "z"], geom.scale)))
                    else:
                        msg.scale = Vector3(x=1.0, y=1.0, z=1.0)
                    msg.mesh_use_embedded_materials = True
                elif isinstance(geom, CylinderVisualShape):
                    msg.type = Marker.CYLINDER
                    msg.scale = Vector3(x=geom.radius * 2, y=geom.radius * 2, z=geom.length)
                elif isinstance(geom, BoxVisualShape):
                    msg.type = Marker.CUBE
                    size = np.array(geom.size) * 2
                    msg.scale = Vector3(x=float(size[0]), y=float(size[1]), z=float(size[2]))
                elif isinstance(geom, SphereVisualShape):
                    msg.type = Marker.SPHERE
                    msg.scale = Vector3(x=geom.radius * 2, y=geom.radius * 2, z=geom.radius * 2)

                marker_array.markers.append(msg)
        return marker_array

    def _stop_publishing(self) -> None:
        """
        Stops the publishing of the Visualization Marker update by setting the kill event and collecting the thread.
        """
        self.kill_event.set()
        self.thread.join()


class ManualMarkerPublisher:
    """
    Class to manually add and remove marker of objects and poses.
    """

    def __init__(self, topic_name: str = '/pycram/manual_marker', interval: float = 0.1):
        """
        The Publisher creates an Array of Visualization marker with a marker for a pose or object.
        This Array is published with a rate of interval.

        :param topic_name: Name of the marker topic
        :param interval: Interval at which the marker should be published
        """
        self.start_time = None
        self.marker_array_pub = create_publisher(topic_name, MarkerArray, queue_size=10)

        self.marker_array = MarkerArray()
        self.marker_overview = {}
        self.current_id = 0

        self.interval = interval
        self.log_message = None

    def publish(self, pose: PoseStamped, color: Optional[List] = None, bw_object: Optional[ObjectDesignatorDescription] = None,
                name: Optional[str] = None):
        """
        Publish a pose or an object into the MarkerArray.
        Priorities to add an object if possible

        :param pose: Pose of the marker
        :param color: Color of the marker if no object is given
        :param bw_object: Object to add as a marker
        :param name: Name of the marker
        """

        if color is None:
            color = [1, 0, 1, 1]

        self.start_time = time.time()
        thread = threading.Thread(target=self._publish, args=(pose, bw_object, name, color))
        thread.start()
        loginfo(self.log_message)
        thread.join()

    def _publish(self, pose: PoseStamped, bw_object: Optional[ObjectDesignatorDescription] = None, name: Optional[str] = None,
                 color: Optional[List] = None):
        """
        Publish the marker into the MarkerArray
        """
        stop_thread = False
        duration = 2

        while not stop_thread:
            if time.time() - self.start_time > duration:
                stop_thread = True
            if bw_object is None:
                self._publish_pose(name=name, pose=pose, color=color)
            else:
                self._publish_object(name=name, pose=pose, bw_object=bw_object)

            sleep(self.interval)

    def _publish_pose(self, name: str, pose: PoseStamped, color: Optional[List] = None):
        """
        Publish a Pose as a marker

        :param name: Name of the marker
        :param pose: Pose of the marker
        :param color: Color of the marker
        """

        if name is None:
            name = 'pose_marker'

        if name in self.marker_overview.keys():
            self._update_marker(self.marker_overview[name], new_pose=pose)
            return

        color_rgba = ColorRGBA(**dict(zip(["r", "g", "b","a"], color)))
        self._make_marker_array(name=name, marker_type=Marker.ARROW, marker_pose=pose,
                                marker_scales=(0.05, 0.05, 0.05), color_rgba=color_rgba)
        self.marker_array_pub.publish(self.marker_array)
        self.log_message = f"Pose '{name}' published"

    def _publish_object(self, name: Optional[str], pose: PoseStamped, bw_object: ObjectDesignatorDescription):
        """
        Publish an Object as a marker

        :param name: Name of the marker
        :param pose: Pose of the marker
        :param bw_object: ObjectDesignatorDescription for the marker
        """

        bw_real = bw_object.resolve()

        if name is None:
            name = bw_real.name

        if name in self.marker_overview.keys():
            self._update_marker(self.marker_overview[name], new_pose=pose)
            return

        path = bw_real.world_object.root_link.geometry.file_name

        self._make_marker_array(name=name, marker_type=Marker.MESH_RESOURCE, marker_pose=pose,
                                path_to_resource=path)

        self.marker_array_pub.publish(self.marker_array)
        self.log_message = f"Object '{name}' published"

    def _make_marker_array(self, name, marker_type: int, marker_pose: PoseStamped, marker_scales: Tuple = (1.0, 1.0, 1.0),
                           color_rgba: ColorRGBA = ColorRGBA(**dict(zip(["r", "g", "b","a"], [1.0, 1.0, 1.0, 1.0]))),
                           path_to_resource: Optional[str] = None):
        """
        Create a Marker and add it to the MarkerArray

        :param name: Name of the Marker
        :param marker_type: Type of the marker to create
        :param marker_pose: Pose of the marker
        :param marker_scales: individual scaling of the markers axes
        :param color_rgba: Color of the marker as RGBA
        :param path_to_resource: Path to the resource of a Bulletworld object
        """

        frame_id = marker_pose.header.frame_id
        new_marker = Marker()
        new_marker.id = self.current_id
        new_marker.header.frame_id = frame_id
        new_marker.ns = name
        # new_marker.header.stamp = Time.now()
        new_marker.type = marker_type
        new_marker.action = Marker.ADD
        new_marker.pose = marker_pose.pose
        new_marker.scale.x = marker_scales[0]
        new_marker.scale.y = marker_scales[1]
        new_marker.scale.z = marker_scales[2]
        new_marker.color.a = color_rgba.a
        new_marker.color.r = color_rgba.r
        new_marker.color.g = color_rgba.g
        new_marker.color.b = color_rgba.b

        if path_to_resource is not None:
            new_marker.mesh_resource = 'file://' + path_to_resource

        self.marker_array.markers.append(new_marker)
        self.marker_overview[name] = new_marker.id
        self.current_id += 1

    def _update_marker(self, marker_id: int, new_pose: PoseStamped) -> bool:
        """
        Update an existing marker to a new pose

        :param marker_id: id of the marker that should be updated
        :param new_pose: Pose where the updated marker is set

        :return: True if update was successful, False otherwise
        """

        # Find the marker with the specified ID
        for marker in self.marker_array.markers:
            if marker.id == marker_id:
                # Update successful
                marker.pose = new_pose
                self.log_message = f"Marker '{marker.ns}' updated"
                self.marker_array_pub.publish(self.marker_array)
                return True

        # Update was not successful
        logwarn(f"Marker {marker_id} not found for update")
        return False

    def remove_marker(self, bw_object: Optional[ObjectDesignatorDescription] = None, name: Optional[str] = None):
        """
        Remove a marker by object or name

        :param bw_object: Object which marker should be removed
        :param name: Name of object that should be removed
        """

        if bw_object is not None:
            bw_real = bw_object.resolve()
            name = bw_real.name

        if name is None:
            logerr('No name for object given, cannot remove marker')
            return

        marker_id = self.marker_overview.pop(name)

        for marker in self.marker_array.markers:
            if marker.id == marker_id:
                marker.action = Marker.DELETE

        self.marker_array_pub.publish(self.marker_array)
        self.marker_array.markers.pop(marker_id)

        loginfo(f"Removed Marker '{name}'")

    def clear_all_marker(self):
        """
        Clear all existing markers
        """
        for marker in self.marker_array.markers:
            marker.action = Marker.DELETE

        self.marker_overview = {}
        self.marker_array_pub.publish(self.marker_array)

        loginfo('Removed all markers')



class TrajectoryPublisher:
    """
    Publishes a trajectory as a MarkerArray to visualize it in rviz.
    """

    @cached_property
    def publisher(self):
        pub = create_publisher("/pycram/trajectory", MarkerArray)
        time.sleep(0.5) # this is needed to synchronize the publisher creation thread
        return pub

    def visualize_trajectory(self, trajectory: List[PoseStamped]):
        """
        Visualize a trajectory in rviz as a series of arrows.

        :param trajectory: The trajectory to visualize as a list of points where an arrow should be drawn
        from point to point.
        """
        marker_array = MarkerArray()
        for index, (p1, p2) in enumerate(zip(trajectory, trajectory[1:])):

            marker = Marker()
            marker.header.frame_id = p1.frame_id
            marker.id = index
            marker.ns = "trajectory_arrows"
            marker.action = Marker.ADD
            marker.type = Marker.ARROW
            marker.lifetime = Duration(60)

            marker_p1 = Point()
            marker_p1.x = p1.position.x
            marker_p1.y = p1.position.y
            marker_p1.z = p1.position.z

            marker_p2 = Point()
            marker_p2.x = p2.position.x
            marker_p2.y = p2.position.y
            marker_p2.z = p2.position.z

            marker.points.append(marker_p1)
            marker.points.append(marker_p2)

            marker.scale.x = 0.01
            marker.scale.y = 0.02
            marker.scale.z = 0.02

            marker.color.r = 1.0
            marker.color.g = 0.0
            marker.color.b = 1.0
            marker.color.a = 1.0

            marker_array.markers.append(marker)
        self.publisher.publish(marker_array)




from ..datastructures.dataclasses import ManipulatorData
from ..datastructures.enums import Grasp, Arms
from ..helper import get_robot_urdf_and_mjcf_file_paths, find_multiverse_resources_path
from ..robot_description import RobotDescriptionManager, create_manipulator_description
from ..ros import logwarn, loginfo
from ..units import meter

data = ManipulatorData(
    name="panda",
    relative_dir="franka_emika_panda",
    base_link="panda",

    arm_end_link="link7",
    joint_names=[f'joint{i}' for i in range(1, 8)],
    home_joint_values=[0, 0, 0, -1.57079, 0, 1.57079, -0.7853],

    gripper_name="hand",
    gripper_relative_dir=None,
    gripper_tool_frame="right_finger",

    gripper_joint_names=[f'finger_joint{i}' for i in [1, 2]],
    closed_joint_values=[0.0, 0.0],
    open_joint_values=[0.04, 0.04],
    opening_distance=0.08 * meter,
    fingers_link_names=["left_finger", "right_finger"])

multiverse_resources = find_multiverse_resources_path()

urdf_filename, mjcf_filename = None, None
if multiverse_resources is not None:
    urdf_filename, mjcf_filename = get_robot_urdf_and_mjcf_file_paths(data.name, data.relative_dir,
                                                                      multiverse_resources=multiverse_resources)

if mjcf_filename is None or urdf_filename is None:
    loginfo(f"Could not initialize {data.name} description as Multiverse resources path not found.")
else:
    robot_description = create_manipulator_description(data, urdf_filename, mjcf_filename)
    ################################# Grasps ##################################
    robot_description.get_arm_chain(Arms.RIGHT).end_effector.update_all_grasp_orientations([0, 0, 0, 1])

    # Add to RobotDescriptionManager
    rdm = RobotDescriptionManager()
    rdm.register_description(robot_description)



import copy
import itertools
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Any, Type, List

import numpy as np
from probabilistic_model.probabilistic_circuit.nx.helper import fully_factorized
from probabilistic_model.probabilistic_circuit.nx.probabilistic_circuit import ProbabilisticCircuit
from probabilistic_model.probabilistic_model import ProbabilisticModel
from random_events.product_algebra import Event, SimpleEvent
from random_events.set import Set
from random_events.variable import Symbolic, Integer, Variable, Continuous

from .datastructures.partial_designator import PartialDesignator
from .language import SequentialPlan
from .plan import Plan, DesignatorNode, ActionNode


@dataclass
class Parameterizer:
    """
    Class that generates possible parameters for a plan.

    This class can be used to generate possible parameters for a plan from probability distributions.
    """

    plan: Plan
    """
    The plan to generate the parameters for.
    """

    parameters: Dict[DesignatorNode, Any] = field(default_factory=dict, init=False)
    """
    A dictionary that maps all nodes in the plan that hold actions to their parameters.
    """

    variables_of_node: Dict[DesignatorNode, List[Variable]] = field(default_factory=dict, init=False)
    """
    A dictionary that maps all nodes in the plan that hold actions to the variables that describe that nodes parameters.
    """

    def __post_init__(self):
        self.make_parameters()
        self.make_variables()

    @property
    def variables(self) -> List[Variable]:
        """
        :return: The variables for all parameters in the plan.
        """
        return list(itertools.chain.from_iterable(self.variables_of_node.values()))

    def get_variable(self, name: str) -> Variable:
        """
        :param name: The name of the variable.
        :return: The variable.
        """
        return [v for v  in self.variables if v.name == name][0]

    def make_parameters(self):
        """
        Create the parameters for all relevant nodes in the plan.
        """
        for node in self.plan.nodes:
            if isinstance(node, DesignatorNode):
                self.parameters[node] = node.action._parameters

    def make_variables(self):
        """
        Create the variables for all relevant parameters in the plan.
        """
        for index, (node, params) in enumerate(self.parameters.items()):
            variables = []
            for name, leaf_type in node.flattened_parameters().items():
                full_name = f"{node.action.__name__}_{index}.{name}"
                variable = leaf_type_to_variable(full_name, leaf_type)
                variables.append(variable)
            self.variables_of_node[node] = variables


    def plan_from_sample(self, model: ProbabilisticModel, sample: np.array) -> Plan:
        """
        Create a sequential plan from a sample of all parameters.

        :param model: The model that generated the sample.
        :param sample: The sample to generate the plan from.
        :return: The executable, sequential plan
        """
        sub_plans = []

        for node in self.variables_of_node:
            flattened_parameters = []
            for variable in self.variables_of_node[node]:
                flattened_parameters.append(sample[model.variables.index(variable)])
            resolved = node.action.reconstruct(flattened_parameters)

            kwargs = {key: getattr(resolved, key) for key in resolved._parameters}
            sub_plans.append(Plan(ActionNode(designator_ref=PartialDesignator(node.action, **kwargs),
                       action = node.action, kwargs=kwargs)))

        return SequentialPlan(*sub_plans)

    def create_fully_factorized_distribution(self) -> ProbabilisticCircuit:
        """
        :return: a fully factorized distribution for the plan.
        """
        distribution = fully_factorized(self.variables, means={v: 0 for v in self.variables},
                                        variances={v: 1 for v in self.variables})
        return distribution

    def create_restrictions(self) -> SimpleEvent:
        """
        :return: The restrictions present in the plan as random event.
        """
        restrictions = {}
        for index, (node, variables) in enumerate(self.variables_of_node.items()):
            for variable in variables:
                parameter_name = variable.name.split(".", 1)[1]
                restriction = node.kwargs.get(parameter_name, None)

                if restriction is None:
                    continue

                restrictions[variable] = restriction

        # create the restrictions as random event
        result = SimpleEvent(restrictions)
        result.fill_missing_variables(self.variables)
        return result


def leaf_type_to_variable(name: str, leaf_type: Type) -> Variable:
    """
    Convert a leaf type to a random events variable.
    :param name: The name of the variable.
    :param leaf_type: The leaf type.
    :return: The variable.
    """
    if issubclass(leaf_type, bool):
        return Symbolic(name, Set.from_iterable([True, False]))
    elif issubclass(leaf_type, Enum):
        return Symbolic(name, Set.from_iterable(leaf_type))
    elif issubclass(leaf_type, int):
        return Integer(name)
    elif issubclass(leaf_type, float):
        return Continuous(name)
    else:
        raise NotImplementedError(f"No conversion between {leaf_type} and random_events.Variable is known.")



import unittest

import numpy as np
from owlready2 import Imp

from pycram.tf_transformations import quaternion_from_euler

from pycram.datastructures.dataclasses import Color, BoundingBox as BB
from pycram.datastructures.pose import PoseStamped
from pycram.testing import BulletWorldTestCase
from pycrap.ontologies import Room, Food


@unittest.skip
class TestLinks(BulletWorldTestCase):

    def tearDown(self):
        super().tearDown()
        self.world.reset_concepts()

    def test_automatic_containment_detection(self):

        self.world.update_containment_for([self.milk])

        self.assertFalse(self.kitchen.contains_body(self.kitchen.links["iai_fridge_main"]))
        self.assertFalse(self.kitchen.contains_body(self.milk))

        self.milk.set_position(self.kitchen.links["iai_fridge_main"].position)

        self.world.update_containment_for([self.milk])

        self.assertTrue(self.kitchen.links["iai_fridge_main"].contains_body(self.milk))

    def test_automatic_containment_detection_from_parts(self):

        self.assertFalse(self.kitchen.contains_body(self.kitchen.links["iai_fridge_main"]))
        self.assertFalse(self.kitchen.contains_body(self.milk))

        self.world.update_containment_for([self.kitchen])
        self.world.update_containment_for([self.milk])

        self.milk.set_position(self.kitchen.links["iai_fridge_main"].position)

        self.world.update_containment_for([self.milk])

        self.assertTrue(self.kitchen.contains_body(self.milk))
        self.assertTrue(self.kitchen.links["iai_fridge_main"].contains_body(self.milk))

    def test_contains_body(self):

        self.assertFalse(self.kitchen.contains_body(self.kitchen.links["iai_fridge_main"]))
        self.assertFalse(self.kitchen.contains_body(self.milk))

        self.kitchen.contained_bodies = [self.kitchen.links["iai_fridge_main"]]
        self.assertTrue(self.kitchen.contains_body(self.kitchen.links["iai_fridge_main"]))
        self.assertFalse(self.kitchen.contains_body(self.milk))

        self.kitchen.links["iai_fridge_main"].contained_bodies = [self.milk]

        # Kitchen should contain cabinet and drawer (reasoned)
        self.assertTrue(self.kitchen.contains_body(self.kitchen.links["iai_fridge_main"]))
        self.assertTrue(self.kitchen.links["iai_fridge_main"].contains_body(self.milk))
        self.assertTrue(self.kitchen.contains_body(self.milk))
        self.assertEqual(len(self.kitchen.contained_bodies), 2)

    def test_get_convex_hull(self):
        self.milk.set_orientation(quaternion_from_euler(0, np.pi / 4, 0))
        hull = self.milk.root_link.get_convex_hull()
        self.assertIsNotNone(hull)
        self.assertTrue(len(hull.vertices) > 0)
        self.assertTrue(len(hull.faces) > 0)
        plot = False
        if plot:
            BB.plot_3d_points([hull.vertices])

    def test_rotated_bounding_box(self):
        self.milk.set_pose(PoseStamped.from_list([1, 1, 1], quaternion_from_euler(np.pi / 4, 0, 0)))
        aabb = self.milk.get_axis_aligned_bounding_box()
        aabb_points = np.array(aabb.get_points_list())
        rbb = self.milk.get_rotated_bounding_box()
        rot_points = np.array(rbb.get_points_list())
        plot = False
        if plot:
            BB.plot_3d_points([aabb_points, rot_points])

    def test_add_constraint(self):
        milk_link = self.milk.root_link
        robot_link = self.robot.root_link
        robot_link.add_fixed_constraint_with_link(milk_link)
        self.assertTrue(milk_link in robot_link.constraint_ids)
        self.assertTrue(robot_link in milk_link.constraint_ids)

    def test_remove_constraint(self):
        milk_link = self.milk.root_link
        robot_link = self.robot.root_link
        robot_link.add_fixed_constraint_with_link(milk_link)
        robot_link.remove_constraint_with_link(milk_link)
        self.assertTrue(milk_link not in robot_link.constraint_ids)
        self.assertTrue(robot_link not in milk_link.constraint_ids)

    def test_transform_link(self):
        milk_link = self.milk.root_link
        robot_link = self.robot.root_link
        milk_to_robot = milk_link.get_transform_to_link(robot_link)
        robot_to_milk = robot_link.get_transform_to_link(milk_link)
        self.assertAlmostEqual(robot_to_milk, ~milk_to_robot)

    def test_set_color(self):
        link = self.robot.get_link('base_link')
        link.color = Color(1, 0, 0, 1)
        self.assertEqual(link.color.get_rgba(), [1, 0, 0, 1])



from .dependencies import *


class is_part_of(BaseProperty):
    """
    A relation between any entities, e.g. 'brain is a part of the human body'. See dul:hasPart for additional documentation.
    """


class has_part(BaseProperty):
    """
    A schematic relation between any entities, e.g. 'the human body has a brain as part', '20th century contains year 1923', 'World War II includes the Pearl Harbour event'.

    Parthood should assume the basic properties of mereology: transitivity, antisymmetry, and reflexivity (propert Parthood of course misses reflexivity).
    However, antisymmetry is not supported in OWL2 explicitly, therefore DUL has to adopt one of two patterns:
    1) dropping asymmetry axioms, while granting reflexivity: this means that symmetry is not enforced, but permitted for the case of reflexivity. Of course, in this way we cannot prevent symmetric usages of hasPart;
    2) dropping the reflexivity axiom, and enforce asymmetry: in this case, we would prevent all symmetric usages, but we loose the possibility of enforcing reflexivity, which is commonsensical in parthood.
    In DUL, we adopt pattern #1 for partOf, and pattern #2 for properPartOf, which seems a good approximation: due to the lack of inheritance of property characteristics, each asymmetric hasPropertPart assertion would also be a reflexive hasPart assertion (reflexive reduction design pattern).

    Subproperties and restrictions can be used to specialize hasPart for objects, events, etc.
    """


class has_parent_link(BaseProperty):
    """
    Relates a joint to the link it connects which is closer to the root of the kinematic chain.
    """


class has_child_link(BaseProperty):
    """
    Relates a joint to the link it connects which is closer to the end of the kinematic chain.
    """


class contains(BaseProperty):
    """
    A schematic relation asserting containment, understood in a very broad sense, by one Entity of another.
    The relation is defined with domain and range of maximum generality, as it is possible to construe containment to
    apply between Events, between Objects, between Qualities and so on. Care should be taken when using it that the
    construal of containment makes sense and is useful. If a clearer relation expresses the connection between two
    Entities, use that relation instead. For example, rather than saying an Event contains an Object, it is probably
    better to say the Event has that Object as a participant. More specific versions of this relation exist,
    e.g. containsEvent, so it is likely that there will be few situations where it should be used itself.
    However, by acting as a superproperty to several relations, it captures a core similarity between these and enables
    taxonomy-based similarity metrics.
    """


class contains_object(BaseProperty):
    """
    A spatial relation holding between a container, and objects it contains.
    """


class is_contained_in(BaseProperty):
    """
    The inverse of the contains relation. See the contains relation for details.
    """

class has_preferred_alignment(BaseProperty):
    """
    A relation between an object and an alignment that is preferred for that object.
    """

class has_preferred_axis(BaseProperty):
    """
    A relation between an object and an axis identifier.
    """

class has_vertical_alignment(BaseProperty):
    """
    A relation between an object and boolean value indicating if the object should be grasped with a vertical alignment.
    """

class has_rotated_gripper(BaseProperty):
    """
    A relation between an object and boolean value indicating if the gripper should be rotated by 90°.
    """

class has_rim_grasp(BaseProperty):
    """
    A relation between an object and a boolean value indicating if the object should be grasped by the rim.
    """


class is_physically_contained_in(BaseProperty):
    """
    A spatial relation holding between an object (the container), and objects it contains.
    """

class supports(BaseProperty):
    """
    A relation between an object (the supporter) and another object (the supportee) where the supporter cancels the
     effect of gravity on the supportee.
    """

class is_supported_by(BaseProperty):
    """
    A relation between an object (the supporter) and another object (the supportee) where the supporter cancels the
     effect of gravity on the supportee. The inverse of the supports relation.
    """


import time

import pycram.world_reasoning as btr
from pycram.testing import BulletWorldTestCase
from pycram.datastructures.pose import PoseStamped
from pycram.robot_description import RobotDescription


class TestCaseBulletWorldReasoning(BulletWorldTestCase):

    def test_contact(self):
        self.milk.set_pose(PoseStamped.from_list([1, 1, 1]))
        self.cereal.set_pose(PoseStamped.from_list([1, 1, 1]))
        time.sleep(1)
        self.assertTrue(btr.contact(self.milk, self.cereal))

    def test_visible(self):
        self.milk.set_pose(PoseStamped.from_list([1.5, 0, 1.2]))
        self.robot.set_pose(PoseStamped.from_list())
        time.sleep(1)
        camera_link = RobotDescription.current_robot_description.get_camera_link()
        self.world.add_vis_axis(self.robot.get_link_pose(camera_link))
        self.assertTrue(btr.visible(self.milk, self.robot.get_link_pose(camera_link),
                                    RobotDescription.current_robot_description.get_default_camera().front_facing_axis))

    def test_occluding(self):
        self.milk.set_pose(PoseStamped.from_list([3, 0, 1.2]))
        self.robot.set_pose(PoseStamped())
        self.assertTrue(btr.occluding(self.milk, self.robot.get_link_pose(
            RobotDescription.current_robot_description.get_camera_link()),
                                      RobotDescription.current_robot_description.get_default_camera().front_facing_axis) != [])

    def test_reachable(self):
        self.robot.set_pose(PoseStamped())
        time.sleep(1)
        self.assertTrue(btr.reachable(PoseStamped.from_list([0.5, -0.5, 0.5]), self.robot,
                                      RobotDescription.current_robot_description.kinematic_chains[
                                          "right"].get_tool_frame()))
        self.assertFalse(btr.reachable(PoseStamped.from_list([2, 2, 1]), self.robot,
                                       RobotDescription.current_robot_description.kinematic_chains[
                                           "right"].get_tool_frame()))


    def test_blocking(self):
        self.milk.set_pose(PoseStamped.from_list([0.5, -0.5, 0.7]))
        self.robot.set_pose(PoseStamped())
        time.sleep(2)
        blocking = btr.blocking(PoseStamped.from_list([0.5, -0.5, 0.7]), self.robot,
                                RobotDescription.current_robot_description.kinematic_chains["right"])
        self.assertTrue(blocking != [])

    def test_supporting(self):
        self.milk.set_pose(PoseStamped.from_list([1.3, 0, 0.9]))
        self.assertTrue(btr.supporting(self.kitchen, self.milk))



import itertools

import networkx as nx
import numpy as np
import plotly.graph_objects as go
from random_events.interval import SimpleInterval, Bound, singleton
from random_events.product_algebra import SimpleEvent, Event
from sortedcontainers import SortedSet
from tqdm import tqdm
from typing_extensions import Self, Optional, List

from .datastructures.world import World
from .datastructures.dataclasses import BoundingBox
from .datastructures.pose import PoseStamped
from .failures import PlanFailure
from .ros import loginfo
try:
    from .ros_utils.viz_marker_publisher import TrajectoryPublisher
except ImportError:
    loginfo("Could not import TrajectoryPublisher. This is probably because you are not running ROS.")

class PoseOccupiedError(PlanFailure):
    """
    Error that is raised when a pose is occupied or not in the search space of a Connectivity Graphs.
    """

    def __init__(self, pose: PoseStamped):
        """
        Construct a new pose occupied error.
        :param pose: The pose that is occupied.
        """
        super().__init__(f"The pose {pose} is occupied.")
        self.pose = pose


class GraphOfConvexSets(nx.Graph):
    """
    A graph that represents the connectivity between convex sets.

    Every node in this graph represents a convex set.
    Every edge represents an adjacency between two convex sets.
    Furthermore, the adjacency is saved in and edge attribute called "intersection".
    """

    search_space: BoundingBox
    """
    The bounding box of the search space. Defaults to the entire three dimensional space.
    """

    def __init__(self, search_space: Optional[BoundingBox] = None):
        super().__init__()
        self.search_space = self._make_search_space(search_space)

    def calculate_connectivity(self, tolerance=.001):
        """
        Generate the edges for this graph.

        :param tolerance: The tolerance for the intersection.
        This value enlarges the boxes before the intersection to account for numeric imprecision.
        """

        number_of_nodes = len(self.nodes)

        # make the bounding boxes a bit larger
        [node.enlarge_all(tolerance) for node in self.nodes]

        def check_connectivity(node1: BoundingBox, node2: BoundingBox):
            """
            Check the connectivity between two nodes.
            Add an edge if they are connected.
            :param node1: The first node.
            :param node2: The second node.
            """
            intersection = node1.intersection_with(node2)
            if intersection:
                intersection.enlarge_all(-tolerance)
                self.add_edge(node1, node2, intersection=intersection)

        # calculate the connectivity for each edge pair
        [check_connectivity(n1, n2) for n1, n2 in tqdm(itertools.combinations(self.nodes, 2),
                                                       desc="Calculating connectivity",
                                                       total=number_of_nodes * (number_of_nodes - 1) // 2)]

        # recreate the original bounding boxes
        [node.enlarge_all(-tolerance) for node in self.nodes]

    def plot_free_space(self) -> List[go.Mesh3d]:
        """
        Plot the free space of the environment in blue.
        :return: A list of traces that can be put into a plotly figure.
        """
        free_space = Event(*[node.simple_event for node in self.nodes])
        return free_space.plot(color="blue")

    def plot_occupied_space(self) -> List[go.Mesh3d]:
        """
        Plot the occupied space of the environment in red.
        :return: A list of traces that can be put into a plotly figure.
        """
        free_space = Event(*[node.simple_event for node in self.nodes])
        occupied_space = ~free_space & self.search_space.simple_event.as_composite_set()
        return occupied_space.plot(color="red")

    def node_of_pose(self, x, y, z) -> Optional[BoundingBox]:
        """
        Find the node that contains a point.

        :param x: The x-coordinate.
        :param y: The y-coordinate.
        :param z: The z-coordinate.
        :return: The node that contains the point or None if no node contains the point.
        """
        for node in self.nodes:
            if node.contains(x, y, z):
                return node
        return None

    def path_from_to(self, start: PoseStamped, goal: PoseStamped) -> Optional[List[PoseStamped]]:
        """
        Calculate a connected path from a start pose to a goal pose.

        :param start: The start pose.
        :param goal: The goal pose.
        :return: The path as a sequence of points to navigate to or None if no path exists.
        """

        # get poses from params
        start_node = self.node_of_pose(*start.position.to_list())
        goal_node = self.node_of_pose(*goal.position.to_list())

        # validate if the poses are part of the graph
        if start_node is None:
            raise PoseOccupiedError(start)
        if goal_node is None:
            raise PoseOccupiedError(goal)

        if start_node == goal_node:
            return [start, goal]

        # get the shortest path (perhaps replace with a*?)
        path = nx.shortest_path(self, start_node, goal_node)

        # if it is not possible to find a path
        if len(path) == 0:
            return None

        # build the path
        result = [start]

        for source, target in zip(path, path[1:]):
            intersection: BoundingBox = self[source][target]["intersection"]
            x_target = intersection.x_interval.center()
            y_target = intersection.y_interval.center()
            z_target = intersection.z_interval.center()
            result.append(PoseStamped.from_list([x_target, y_target, z_target], [0, 0, 0, 1]))

        result.append(goal)
        return result

    @classmethod
    def _make_search_space(cls, search_space: Optional[BoundingBox] = None):
        """
        Create the default search space if it is not given.
        """
        if search_space is None:
            search_space = BoundingBox(-np.inf, -np.inf, -np.inf,
                                       np.inf, np.inf, np.inf)
        return search_space

    @classmethod
    def obstacles_of_world(cls, world: World, search_space: Optional[BoundingBox] = None) -> Event:
        """
        Get all obstacles of the world besides the robot as a random event.
        """

        # create search space for calculations
        search_space = cls._make_search_space(search_space)
        search_event = search_space.simple_event.as_composite_set()

        # initialize obstacles
        obstacles = None

        # create an event that describes the obstacles in the belief state of the robot
        for obj in world.objects:

            # don't take self-collision into account
            if obj.is_a_robot:
                continue

            # get the bounding box of every link in the object description
            for link in obj.link_name_to_id.keys():
                bb = obj.get_link_axis_aligned_bounding_box(link)

                # limit bounding box to search space
                bb_event = bb.simple_event.as_composite_set() & search_event

                # skip bounding boxes that are outside the search space
                if bb_event.is_empty():
                    continue

                # update obstacles
                if obstacles is None:
                    obstacles = bb_event
                else:
                    obstacles |= bb_event

        return obstacles

    @classmethod
    def free_space_from_world(cls, world: World, tolerance=.001, search_space: Optional[BoundingBox] = None) -> Self:
        """
        Create a connectivity graph from the free space in the belief state of the robot.

        :param world: The belief state.
        :param tolerance: The tolerance for the intersection when calculating the connectivity.
        :param search_space: The search space for the connectivity graph.
        :return: The connectivity graph.
        """

        # create search space for calculations
        search_space = cls._make_search_space(search_space)
        search_event = search_space.simple_event.as_composite_set()

        # get obstacles
        obstacles = cls.obstacles_of_world(world, search_space)

        # calculate the free space and limit it to the searching space
        free_space = ~obstacles & search_event

        # create a connectivity graph from the free space and calculate the edges
        result = cls(search_space=search_space)
        result.add_nodes_from(BoundingBox.from_event(free_space))
        result.calculate_connectivity(tolerance)

        return result

    @classmethod
    def navigation_map_from_world(cls, world: World, tolerance=.001,
                                  search_space: Optional[BoundingBox] = None) -> Self:
        """
        Create a GCS from the free space in the belief state of the robot for navigation.
        The resulting GCS describes the paths for navigation, meaning that changing the z-axis position is not
        possible.
        Furthermore, it is taken into account that the robot has to fit through the entire space and not just
        through the floor level obstacles.

        TODO: i think explicitly 2d boxes would help here.

        :param world: The belief state.
        :param tolerance: The tolerance for the intersection when calculating the connectivity.
        :param search_space: The search space for the connectivity graph.
        :return: The connectivity graph.
        """

        xy = SortedSet([BoundingBox.x_variable, BoundingBox.y_variable])

        # create search space for calculations
        search_space = cls._make_search_space(search_space)

        # remove the z axis
        search_event = search_space.simple_event.as_composite_set()
        search_event = search_event.marginal(xy)

        # get obstacles
        obstacles = cls.obstacles_of_world(world, search_space)

        # get the 2d map from the obstacles
        obstacles = obstacles.marginal(xy)
        free_space = ~obstacles & search_event

        # create floor level
        z_event = SimpleEvent({BoundingBox.z_variable: singleton(0.)}).as_composite_set()
        z_event.fill_missing_variables(xy)
        free_space.fill_missing_variables(SortedSet([BoundingBox.z_variable]))
        free_space &= z_event

        # update z for search space
        search_space.z = SimpleInterval(0., 0., Bound.CLOSED, Bound.CLOSED)

        # create a connectivity graph from the free space and calculate the edges
        result = cls(search_space=search_space)
        result.add_nodes_from(BoundingBox.from_event(free_space))
        result.calculate_connectivity(tolerance)

        return result


def plot_path_in_rviz(path: List[PoseStamped]):
    """
    Plot a path in rviz.

    :param path: The path to plot.
    """

    def make_publisher():
        return TrajectoryPublisher()

    publisher = make_publisher()
    publisher.visualize_trajectory(path)



from typing_extensions import Optional

from ..datastructures.enums import GripperState
from ..designators.motion_designator import MoveGripperMotion, TalkingMotion
from ..ros import  loginfo, logwarn
from ..ros import  create_publisher
from ..ros import  Rate

is_init = False


def init_tmc_interface():
    global is_init
    if is_init:
        return
    from tmc_control_msgs.msg import GripperApplyEffortActionGoal
    from tmc_msgs.msg import Voice
    is_init = True
    loginfo("Successfully initialized tmc interface")



def tmc_gripper_control(designator: MoveGripperMotion, topic_name: Optional[str] = '/hsrb/gripper_controller/grasp/goal'):
    """
    Publishes a message to the gripper controller to open or close the gripper for the HSR.

    :param designator: The designator containing the motion to be executed
    :param topic_name: The topic name to publish the message to
    """
    if (designator.motion == GripperState.OPEN):
        pub_gripper = create_publisher(topic_name, GripperApplyEffortActionGoal, 10)
        rate = Rate(10)
        msg = GripperApplyEffortActionGoal()
        msg.goal.effort = 0.8
        pub_gripper.publish(msg)

    elif (designator.motion == GripperState.CLOSE):
        pub_gripper = create_publisher(topic_name, GripperApplyEffortActionGoal, 10)
        rate = Rate(10)
        msg = GripperApplyEffortActionGoal()
        msg.goal.effort = -0.8
        pub_gripper.publish(msg)


def tmc_talk(designator: TalkingMotion, topic_name: Optional[str] = '/talk_request'):
    """
    Publishes a sentence to the talk_request topic of the HSRB robot

    :param designator: The designator containing the sentence to be spoken
    :param topic_name: The topic name to publish the sentence to
    """
    pub = create_publisher(topic_name, Voice, 10)
    texttospeech = Voice()
    # language 1 = english (0 = japanese)
    texttospeech.language = 1
    texttospeech.sentence = designator.cmd

    pub.publish(texttospeech)



from ..ros import   get_ros_package_path
from ..robot_description import RobotDescription, CameraDescription, KinematicChainDescription, \
    EndEffectorDescription, RobotDescriptionManager
from ..datastructures.enums import Arms, Grasp, GripperState, TorsoState, GripperType, StaticJointState
from ..units import meter
from ..datastructures.dataclasses import VirtualMobileBaseJoints

filename = get_ros_package_path('pycram') + '/resources/robots/' + "boxy" + '.urdf'

boxy_description = RobotDescription("boxy", "base_link", "triangle_base_link", "triangle_base_joint",
                                    filename, virtual_mobile_base_joints=VirtualMobileBaseJoints())

################################## Right Arm ##################################
right_arm = KinematicChainDescription("right_arm", "calib_right_arm_base_link", "right_arm_7_link",
                                      boxy_description.urdf_object, arm_type=Arms.RIGHT)

right_arm.add_static_joint_states(StaticJointState.Park, {"right_arm_0_joint": 1.858,
                                           "right_arm_1_joint": -0.70571,
                                           "right_arm_2_joint": -0.9614,
                                           "right_arm_3_joint": 0.602,
                                           "right_arm_4_joint": 2.5922,
                                           "right_arm_5_joint": 1.94065,
                                           "right_arm_6_joint": 1.28735})

boxy_description.add_kinematic_chain_description(right_arm)

################################## Right Gripper ##################################

right_gripper = EndEffectorDescription("right_gripper", "right_gripper_base_link", "right_gripper_tool_frame",
                                       boxy_description.urdf_object)

right_gripper.add_static_joint_states(GripperState.OPEN, {"right_gripper_joint": 0.548})
right_gripper.add_static_joint_states(GripperState.CLOSE, {"right_gripper_joint": 0.0})
right_gripper.end_effector_type = GripperType.PARALLEL  # current gripper in sim, change later
right_gripper.opening_distance = 0.11 * meter  # 2x 55mm for WSG050
right_arm.end_effector = right_gripper

################################## Left Arm ##################################
left_arm = KinematicChainDescription("left_arm", "calib_left_arm_base_link", "left_arm_7_link",
                                     boxy_description.urdf_object, arm_type=Arms.LEFT)

left_arm.add_static_joint_states(StaticJointState.Park, {"left_arm_0_joint": -1.858,
                                          "left_arm_1_joint": 0.70571,
                                          "left_arm_2_joint": 0.9614,
                                          "left_arm_3_joint": -0.602,
                                          "left_arm_4_joint": -2.5922,
                                          "left_arm_5_joint": -1.94065,
                                          "left_arm_6_joint": -1.28735})

boxy_description.add_kinematic_chain_description(left_arm)

################################## Left Gripper ##################################

left_gripper = EndEffectorDescription("left_gripper", "left_gripper_base_link", "left_gripper_tool_frame",
                                      boxy_description.urdf_object)

left_gripper.add_static_joint_states(GripperState.OPEN, {"left_gripper_joint": 0.548})
left_gripper.add_static_joint_states(GripperState.CLOSE, {"left_gripper_joint": 0.0})
left_gripper.end_effector_type = GripperType.PARALLEL  # current gripper in sim, change later
left_gripper.opening_distance = 0.11 * meter  # 2x 55mm for WSG050
left_arm.end_effector = left_gripper

################################## Torso ##################################
torso = KinematicChainDescription("torso", "base_link", "triangle_base_link",
                                  boxy_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"triangle_base_joint": 0})

torso.add_static_joint_states(TorsoState.MID, {"triangle_base_joint": 0.29})

torso.add_static_joint_states(TorsoState.LOW, {"triangle_base_joint": -0.58})

boxy_description.add_kinematic_chain_description(torso)

################################## Camera ##################################
camera = CameraDescription("head_mount_kinect2_rgb_optical_frame", "head_mount_kinect2_rgb_optical_frame",
                           2.5, 0.99483, 0.75049)
boxy_description.add_camera_description(camera)

################################## Neck ##################################
boxy_description.add_kinematic_chain("neck", "neck_base_link", "neck_wrist_3_link")

################################# Grasps ##################################
left_gripper.update_all_grasp_orientations([1, 0, 1, 0])
right_gripper.update_all_grasp_orientations([1, 0, 1, 0])

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(boxy_description)



import shutil
from os import mkdir

import inflection
import networkx as nx
import owlready2
import tqdm
from owlready2 import ThingClass, PropertyClass, get_ontology, LogicalClassConstruct
from owlready2.base import *
from typing_extensions import List, Any

# Mapping of digits to words
digit_map = {
    '0': 'Zero', '1': 'One', '2': 'Two', '3': 'Three', '4': 'Four',
    '5': 'Five', '6': 'Six', '7': 'Seven', '8': 'Eight', '9': 'Nine'
}


def set_explicit_repr_for_logical_operator():
    def explicit_repr(self):
        s = []
        for x in self.Classes:
            if isinstance(x, LogicalClassConstruct):
                s.append("(%s)" % x)
            else:
                s.append(repr(x))
        return "%s([%s])" % (self.__class__.__name__, ", ".join(s))

    LogicalClassConstruct.__repr__ = explicit_repr


def to_snake_case(string: str) -> str:
    """
    Convert a string to snake case.

    :param string: The string to convert.
    :return: The string in snake case.
    """
    return inflection.underscore(string)


def to_camel_case(string: str) -> str:
    """
    Convert a string to camel case.

    :param string: The string to convert.
    :return: The string in camel case.
    """
    return inflection.camelize(string)


def replace_types(string: str) -> str:
    """
    Replace the types in a string with python types

    Example:
    >>> replace_types("array_double__<class 'int'>")
    "float__int"

    # TODO array_double will be converted to float for now
    :param string: The string to convert
    :return: The string with the types replaced
    """
    if "-" in str(string):
        string = str(string).replace("-", "")
    if "array_double" in str(string):
        string = str(string).replace("array_double", "float")
    if "<class 'int'>" in str(string):
        string = str(string).replace("<class 'int'>", "int")
    if "<class 'str'>" in str(string):
        string = str(string).replace("<class 'str'>", "str")
    if "<class 'owlready2.util.normstr'>" in str(string):
        string = str(string).replace("<class 'owlready2.util.normstr'>", "normstr")
    if "<class 'float'>" in str(string):
        string = str(string).replace("<class 'float'>", "float")
    if "<class 'bool'>" in str(string):
        string = str(string).replace("<class 'bool'>", "bool")
    if "<class 'datetime.datetime'>" in str(string):
        string = str(string).replace("<class 'datetime.datetime'>", "datetime.datetime")
    return string


def update_class_names(onto: owlready2.Ontology):
    """
    Update the class names to match python conventions.
    """
    for cls in onto.classes():
        if cls is owlready2.Thing:
            continue
        converted_name = replace_types(cls.name)
        converted_name = to_camel_case(converted_name)
        converted_name = ''.join(digit_map[char] if char.isdigit() else char for char in converted_name)
        type.__setattr__(cls, "_name", converted_name)


def update_property_names(onto: owlready2.Ontology):
    """
    Update the property names to match python conventions of functions and members.
    """
    for prop in onto.properties():
        converted_name = to_snake_case(prop.name)
        converted_name = ''.join(digit_map[char] if char.isdigit() else char for char in converted_name)
        type.__setattr__(prop, "original_name", prop.name)
        type.__setattr__(prop, "_name", converted_name)


class AbstractParser:
    """
    An abstract class for parsing
    """

    current_file: Any
    """
    The current file where contents are written into.
    """

    indentation: int
    """
    The indentation to use for the python file.
    """

    file_extension: str = ".py"
    """
    The file extension for the python file.
    """

    path: str
    """
    The path to write the files of the parsed ontology into.
    """

    def __init__(self, path, indentation: int = 4):
        self.path = path
        self.indentation = indentation

    def apply_indent_to(self, string):
        """
        Indent a statement at the beginning of every new line.

        :param string: The statement.
        :return: The indented string.
        """
        return " " * self.indentation + string.replace('\n', '\n' + ' ' * self.indentation)

    def path_for_file(self, file_name):
        """
        Generate the path for a file.

        Example:
        >>> parser = AbstractParser("/tmp")
        >>> parser.path_for_file("base")
            "/tmp/base.py"

        :param file_name: The file name.
        :return: The path to the file.
        """
        return f"{os.path.join(self.path, file_name)}{self.file_extension}"


class OntologyParser(AbstractParser):
    """
    A class that parses everything from an owlready2 compatible ontology into python files that
    represent the same ontology.

    It will create several files in a directory specified by the constructor:
    - dependencies: A file that the imports from other packages/modules that are needed to define this ontology.
    - classes: A file that contains all classes from the ontology without any restrictions.
    - object_properties: A file the contains all object properties from the ontology without any restrictions.
    - data_properties: A file the contains all data properties from the ontology without any restrictions.
    - restrictions: The restrictions for all classes and properties.
    - individuals: All individuals with the respective properties from the ontology.
    - __init__.py: A package initialization that loads the content of all files and hence sets the restrictions
                   and so on.

    TODO: Labels metadata and SWRL is not parsed
    """

    ontology: owlready2.Ontology
    """
    The ontology to parse.
    """

    classes_file_name: str = "classes"
    """
    The file name where all elements from ontology.classes() are written to.
    """

    object_properties_file_name: str = "object_properties"
    """
    The file name where all elements from ontology.properties() are written to.
    """

    data_properties_file_name: str = "data_properties"
    """
    The file name where all elements from ontology.data_properties() are written to.
    """

    restrictions_file_name: str = "restrictions"
    """
    The file name where all restrictions are written to.
    """

    individuals_file_name: str = "individuals"
    """
    The file name where all individuals are written to.
    """

    dependencies_file_name: str = "dependencies"
    """
    The file name where all dependencies are written to.
    """

    dependencies: List[owlready2.Ontology]
    """
    The other ontologies that have to be imported.
    """

    def __init__(self, ontology: owlready2.Ontology, dependencies: List[owlready2.Ontology], path: str,
                 indentation: int = 4):
        super().__init__(path, indentation)
        self.ontology = ontology
        self.dependencies = dependencies
        # TODO update class and property names here to match python conventions

    def parse(self):
        """
        Parses the ontology into a python file.
        """
        self.create_dependencies()
        self.create_classes()
        self.create_object_properties()
        self.create_data_properties()
        self.create_individuals()
        self.create_restrictions()
        # create swrl rules
        self.create_init()

    def create_dependencies(self):
        self.current_file = open(self.path_for_file(self.dependencies_file_name), "w")
        self.current_file.write("from ..base import *\n")
        for dependency in self.dependencies:
            self.current_file.write(f"from ..{to_snake_case(dependency.name)} import *\n")
        self.current_file.close()

    def create_init(self):
        """
        Create the __init__.py
        """
        self.current_file = open(self.path_for_file("__init__"), "w")
        self.import_classes()
        self.import_properties()
        self.import_individuals()
        self.import_restrictions()
        self.current_file.close()

    def import_restrictions(self):
        """
        Write the import statement that imports restrictions.
        """
        self.current_file.write("from .restrictions import *\n")

    def import_dependencies(self):
        """
        Import from the dependencies.
        """
        self.current_file.write(f"from .{self.dependencies_file_name} import *\n")

    def create_classes(self):
        """
        Create the classes.py
        """
        self.current_file = open(self.path_for_file(self.classes_file_name), "w")
        self.import_dependencies()
        self.current_file.write("\n" * 2)
        classes = list(self.ontology.classes())
        for cls in tqdm.tqdm(classes, desc="Parsing classes"):
            self.parse_class(cls)
        self.current_file.close()

    def create_object_properties(self):
        """
        Create the object_properties.py
        """
        self.current_file = open(self.path_for_file(self.object_properties_file_name), "w")
        self.import_dependencies()
        self.current_file.write("\n" * 2)
        properties = list(self.ontology.object_properties())
        for prop in tqdm.tqdm(properties, desc="Parsing object properties"):
            self.parse_property(prop)
        self.current_file.close()

    def create_data_properties(self):
        """
        Create the data_properties.py
        """
        self.current_file = open(self.path_for_file(self.data_properties_file_name), "w")
        self.import_dependencies()
        self.import_classes()
        self.current_file.write("\n" * 2)
        properties = list(self.ontology.data_properties())
        for prop in tqdm.tqdm(properties, desc="Parsing data properties"):
            self.parse_property(prop)
        self.current_file.close()

    def create_restrictions(self):
        """
        Create the restrictions.py
        """
        self.current_file = open(self.path_for_file(self.restrictions_file_name), "w")
        self.import_dependencies()
        self.import_classes()
        self.import_individuals()
        self.current_file.write("\n" * 2)

        elements = list(self.ontology.classes()) + list(self.ontology.properties())

        for element in tqdm.tqdm(elements, desc="Parsing restrictions"):
            self.parse_restrictions_for(element)
            self.current_file.write("\n")
        self.current_file.close()

    def parse_restrictions_for(self, element):
        """
        Create all restriction for any element of the ontology.
        """
        if isinstance(element, ThingClass):
            self.parse_restrictions_for_class(element)
        elif isinstance(element, PropertyClass):
            self.parse_restrictions_for_property(element)

    def parse_restrictions_for_class(self, cls: ThingClass):
        """
        Create the restrictions for a class.

        :param cls: The class
        """
        # write is_a restrictions
        is_a = self.parse_elements(cls.is_a)
        if is_a:
            is_a = f"{repr(cls)}.is_a = [{replace_types(str(is_a))}]"
            self.current_file.write(is_a)
            self.current_file.write("\n")

        # write equivalent_to restrictions
        self.write_equivalent_to(cls)

    def import_individuals(self):
        """
        Write the import statement that imports individuals.
        """
        self.current_file.write("from .individuals import *\n")

    def parse_restrictions_for_property(self, prop):
        """
        Write all restrictions for a property.
        :param prop: The property
        """
        # write is_a restrictions
        #Some Properties restrictions may not be converted to snake case yet, this function ensures that restrictions of
        #the properties matches the defined properties.
        updated_is_a = []
        for i in prop.is_a:
            if i not in self.ontology.properties():
                for l in self.ontology.properties():
                    if l.name == to_snake_case(i.name):
                        i = l
            updated_is_a.append(i)

        is_a = self.parse_elements(updated_is_a)
        #print(*self.ontology.properties())
        # Skip AnnotationProperties for now
        if not "AnnotationProperty" in is_a:
            if is_a:
                is_a = f"{repr(prop)}.is_a = [{is_a}]"
                self.current_file.write(is_a)
                self.current_file.write("\n")

            # write domain restrictions
            domain_string = self.parse_elements(prop.domain)
            if domain_string:
                domain_string = f"{repr(prop)}.domain = [{domain_string}]"
                self.current_file.write(domain_string)
                self.current_file.write("\n")

            # write range restrictions
            range_string = self.parse_elements(prop.range)
            if range_string:
                range_string = f"{repr(prop)}.range = [{replace_types(str(range_string))}]"
                self.current_file.write(range_string)
                self.current_file.write("\n")

    def create_individuals(self):
        """
        Create all individuals of the ontology.
        """
        self.current_file = open(f"{os.path.join(self.path, self.individuals_file_name)}{self.file_extension}", "w")
        self.import_dependencies()
        self.import_classes()
        self.import_properties()
        self.current_file.write("\n" * 2)

        individuals = list(self.ontology.individuals())
        for individual in tqdm.tqdm(individuals, desc="Parsing individuals"):
            self.parse_individual(individual)
        self.current_file.write("\n")
        for individual in tqdm.tqdm(individuals, desc="Parsing individuals"):
            self.parse_individual_properties(individual)
            if individual.get_properties():
                self.current_file.write("\n")
        self.current_file.close()

    def parse_individual(self, individual: owlready2.Thing):
        """
        Parse the construction of an individual.
        :param individual: The individual.
        """
        self.current_file.write(f"{individual.name} = {repr(individual.__class__)}(namespace = ontology)")
        self.current_file.write("\n")

    def parse_individual_properties(self, individual: owlready2.Thing):
        """
        Parse the properties of an individual.
        :param individual: The individual.
        """
        for prop in individual.get_properties():
            if "original_name" not in dir(prop):
                continue
           #Some properties from individuals may be buggy, so only use it on demand.
#           self.current_file.write(f"{individual.name}.{prop.name} = {individual.__getattr__(prop.original_name)}")
            self.current_file.write("\n")

    def import_classes(self):
        """
        Create the import statement to get everything from the classes.py
        """
        self.current_file.write("from .classes import *\n")

    def import_properties(self):
        """
        Create the import statement to get everything from the properties
        """
        self.current_file.write("from .object_properties import *\n")
        self.current_file.write("from .data_properties import *\n")

    def apply_indent_to(self, string):
        """
        Indent a statement at the beginning of every new line.

        :param string: The statement.
        :return: The indented string.
        """
        return " " * self.indentation + string.replace('\n', '\n' + ' ' * self.indentation)

    def get_docstring(self, cls) -> str:
        """
        Get the docstring for a class.

        :param cls: The class
        :return: The docstring for the class or "..." if no docstring is found.
        """
        docstring = cls.comment
        label = cls.label
        if docstring and label:
            docstring = f"\n".join(docstring)
            label = f"\n".join(label)
            return (f'"""\n'
                    f'label: {label}\n'
                    f'{docstring}\n'
                    f'"""\n')
        if docstring:
            docstring = f"\n".join(docstring)
            return (f'"""\n'
                    f'{docstring}\n'
                    f'"""\n')
        else:
            return "...\n"

    def parse_element(self, element) -> str:
        """
        Parse an element for representation the source code.

        :param element: The element to parse.
        :return: A string representation that can be used in python files.
        """
        return repr(element)

    def parse_elements(self, elements: List) -> str:
        """
        Parse a list of elements from for the representation in the source code.
        An input can be, for instance, the is_a` field of a class.

        :param elements: A list of elements to parse.
        :return: A string representation of the elements.
        """
        return ", ".join(self.parse_element(element) for element in elements)

    def write_docstring(self, cls):
        """
        Write the docstring of a class to the current file.
        :param cls: The class.
        """
        # apply indent to docstring
        docstring = self.get_docstring(cls)
        docstring = self.apply_indent_to(docstring)
        self.current_file.write(docstring)

    def write_equivalent_to(self, cls):
        """
        Write the `equivalent_to` field of a class

        :param cls: The class.
        """
        equivalent_to = self.parse_elements(cls.equivalent_to)
        if equivalent_to:
            equivalent_to = f"{repr(cls)}.equivalent_to = [{equivalent_to}]"
            self.current_file.write(equivalent_to)
            self.current_file.write("\n")

    def write_is_a(self, cls):
        """
        Write the `is_a` field of a class
        :param cls: The class.
        """
        is_a = self.parse_elements(cls.is_a)
        is_a = f"{repr(cls)}is_a = [{is_a}]"
        self.current_file.write(is_a)
        self.current_file.write("\n")

    # TODO: Decide upon a better solution for the hyphen symbol
    def parse_class(self, cls):
        """
        Parse a class without restrictions.
        :param cls: The class.
        """
        # Adding the full iri and base iri as class attributes.
        inherited_classes_sting = "Base"
        fulliri = "full_iri = " + "'" + cls.iri + "'"
        baseiri = "base_iri = " + "'" + cls.iri.split("#")[0] + "'"
        self.current_file.write(f"class {repr(cls)}({inherited_classes_sting}):\n")
        self.write_docstring(cls)
        self.current_file.write(f"\n{self.apply_indent_to(fulliri)}\n")
        #self.current_file.write("\n\n")
        self.current_file.write(f"\n{self.apply_indent_to(baseiri)}\n")
        self.current_file.write("\n\n")

    def parse_property(self, prop):
        """
        Parse a property without restrictions.
        :param prop: The property.
        """
        # Adding the full iri and base iri as property attributes.
        fulliri = "full_iri = " + "'" + prop.iri + "'"
        baseiri = "base_iri = " + "'" + prop.iri.split("#")[0] + "'"
        self.current_file.write(f"class {prop.name}(BaseProperty):\n")
        self.write_docstring(prop)
        self.current_file.write(f"\n{self.apply_indent_to(fulliri)}\n")
        self.current_file.write(f"\n{self.apply_indent_to(baseiri)}\n")
        self.current_file.write("\n\n")


class OntologiesParser(AbstractParser):
    """
    Class that parses multiple ontologies at once.

    The resulting python package has the following form


    path/__init__.py
    path/base.py
    path/ontology1/__init__.py
    path/ontology1/classes.py
    path/ontology1/object_properties.py
    path/ontology1/data_properties.py
    path/ontology1/restrictions.py
    path/ontology1/individuals.py
    path/ontology2/__init__.py
    path/ontology2/classes.py
    ...
    """

    ontologies: List[owlready2.Ontology]
    """
    The ontologies to parse.
    """

    include_imported_ontologies = True
    """
    If True, the imported ontologies are also parsed.
    """

    dependency_graph: nx.DiGraph
    """
    The dependency graph of the ontologies.
    """

    base_file_name: str = "base"
    """
    The file name where the base classes are written to.
    """

    clear_existing: bool = True
    """
    If True, the existing directories are cleared.
    """

    def __init__(self, ontologies: List[owlready2.Ontology], path: str, indentation: int = 4):

        render_func_without_namespace = lambda entity: entity.name
        owlready2.set_render_func(render_func_without_namespace)

        super().__init__(path, indentation)
        self.ontologies = ontologies

        if self.include_imported_ontologies:
            self.ontologies += [imported for onto in self.ontologies for imported in onto.imported_ontologies]

        # make ontologies unique
        self.ontologies = list(set(self.ontologies))
        self.ontologies.sort(key=lambda onto: onto.name)
        self.create_dependency_graph()
        set_explicit_repr_for_logical_operator()

    def create_base(self):
        """
        Create the base file
        """
        self.current_file = open(self.path_for_file(self.base_file_name), "w")
        self.current_file.write(
            "from owlready2 import Thing, ThingClass, ObjectProperty, get_ontology, And, Or, Not, OneOf, Inverse, "
            "normstr, DatatypeProperty, TransitiveProperty, SymmetricProperty, AsymmetricProperty, ReflexiveProperty, "
            "IrreflexiveProperty, datetime \n")
        self.current_file.write("import tempfile\n")
        self.current_file.write("\n" * 2)
        self.current_file.write("ontology_file = tempfile.NamedTemporaryFile()\n")
        self.current_file.write('ontology = get_ontology("file://" + ontology_file.name).load()\n')
        self.current_file.write("\n" * 2)
        self.create_base_class()
        self.create_base_property()
        self.current_file.close()

    def create_base_class(self):
        """
        Create the base class for concepts.
        """
        self.current_file.write("class Base(Thing, metaclass=ThingClass):\n")
        self.current_file.write(self.apply_indent_to("namespace = ontology"))
        self.current_file.write("\n" * 3)

    def create_base_property(self):
        """
        Create the base class for properties.
        """
        self.current_file.write("class BaseProperty(ObjectProperty):\n")
        self.current_file.write(self.apply_indent_to("namespace = ontology"))
        self.current_file.write("\n" * 3)

    def destroy_all_ontologies(self):
        """
        Destroy all ontologies.
        """
        for onto in self.ontologies:
            onto.destroy()

    def create_dependency_graph(self):
        """
        Create the dependency graph of the ontologies.
        """
        self.dependency_graph = nx.DiGraph()
        for onto in self.ontologies:
            self.dependency_graph.add_node(onto)
            for imported in onto.imported_ontologies:
                self.dependency_graph.add_edge(imported, onto)

    def parse(self):
        self.create_base()
        self.create_ontologies()

    def create_init(self):
        """
        Create the __init__.py
        """
        self.current_file = open(self.path_for_file("__init__"), "w")
        self.current_file.write("from .base import *\n")
        for onto in self.ontologies:
            self.current_file.write(f"from .{to_snake_case(onto.name)} import *\n")
        self.current_file.close()

    def create_ontologies(self):
        self.destroy_all_ontologies()

        for onto in nx.topological_sort(self.dependency_graph):
            loaded_onto: owlready2.Ontology = get_ontology(onto.base_iri)
            # Some Ontologies require a local path to be parsed.
            if loaded_onto.base_iri == "http://knowrob.org/kb/urdf.owl#":
                loaded_onto: owlready2.Ontology = get_ontology("https://raw.githubusercontent.com/knowrob/knowrob/refs/heads/dev/owl/URDF.owl")

            if loaded_onto.base_iri == "http://www.ease-crc.org/ont/SUTURO.owl#":
            #    loaded_onto: owlready2.Ontology = get_ontology("https://raw.githubusercontent.com/SUTURO/suturo_knowledge/refs/heads/fallschool/suturo_knowledge/owl/suturo.owl")
                loaded_onto:  owlready2.Ontology = get_ontology("suturo.owl")

            if loaded_onto.base_iri == "http://www.ease-crc.org/ont/meals#":
                loaded_onto: owlready2.Ontology = get_ontology("meals.owl")

            loaded_onto.load()
            update_class_names(loaded_onto)
            update_property_names(loaded_onto)
            # Create the directory for the ontology
            ontology_path = os.path.join(self.path, to_snake_case(loaded_onto.name))

            if self.clear_existing:
                if os.path.exists(ontology_path):
                    shutil.rmtree(ontology_path)

            mkdir(ontology_path)

            # Parse the ontology
            parser = OntologyParser(loaded_onto, list(self.dependency_graph.predecessors(onto)), ontology_path)
            parser.parse()

        # Create the __init__.py
        self.create_init()



import time
import datetime


class Time:
    """
    Class to abstract the ROS2 Time, to make it more consistent with the ROS1 Time class.
    """
    def __init__(self, time=0.0, nsec=0.0):
        self.time = time
        self.nsec = nsec

    @classmethod
    def now(cls):
        return datetime.datetime.now().timestamp()
        #return builtin_interfaces.msg.Time(**dict(zip(["sec", "nanosec"], node.get_clock().now().seconds_nanoseconds())))
        # return cls(*node.get_clock().now().seconds_nanoseconds())
        #return node.get_clock().now()

    def to_sec(self):
        return datetime.datetime.now().timestamp()
        #return self.sec()[0]

def Duration(duration=0.0):
    return duration

def Rate(rate):
    return rate


class ServiceException(Exception):
    """
    Exception class for service exceptions.
    """
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


import sys
from threading import Lock, RLock
from typing import Any

from ..ros import  create_action_client
from ..ros import  logwarn, loginfo, loginfo_once
from ..ros import  get_node_names

from typing_extensions import List, Callable, Optional

from ..datastructures.pose import PoseStamped
from ..designator import ObjectDesignatorDescription

robokudo_found = False

is_init = False
client = None

number_of_par_goals = 0
robokudo_lock = Lock()
robokudo_rlock = RLock()
with robokudo_rlock:
    par_threads = {}
    par_motion_goal = {}


def thread_safe(func: Callable) -> Callable:
    """
    Adds thread safety to a function via a decorator. This uses the robokudo_lock

    :param func: Function that should be thread safe
    :return: A function with thread safety
    """

    def wrapper(*args, **kwargs):
        with robokudo_rlock:
            return func(*args, **kwargs)

    return wrapper


def init_robokudo_interface(func: Callable) -> Callable:
    """
    Checks if the ROS messages are available and if Robokudo is running, if that is the case the interface will be
    initialized.

    :param func: Function this decorator should be wrapping
    :return: A callable function which initializes the interface and then calls the wrapped function
    """

    def wrapper(*args, **kwargs):

        from robokudo_msgs.msg import ObjectDesignator as robokudo_ObjectDesignator
        from robokudo_msgs.msg import QueryAction, QueryGoal, QueryResult
        from geometry_msgs.msg import PointStamped

        robokudo_found = True

        global is_init
        global client
        if is_init and "/robokudo" in get_node_names():
            return func(*args, **kwargs)
        elif is_init and "/robokudo" not in get_node_names():
            logwarn("Robokudo node is not available anymore, could not initialize robokudo interface")
            is_init = False
            return

        if "robokudo_msgs" not in sys.modules:
            logwarn("Could not initialize the Robokudo interface since the robokudo_msgs are not imported")
            return

        if "/robokudo" in get_node_names():
            loginfo_once("Successfully initialized Robokudo interface")
            is_init = True
            client = create_action_client("robokudo/query", QueryAction)
            loginfo("Waiting for action server")
            if client.wait_for_server():
                loginfo("Action server is available")
            else:
                logwarn("Action server is not available")
                is_init = False
                return
        else:
            logwarn("Robokudo is not running, could not initialize Robokudo interface")
            return
        return func(*args, **kwargs)

    return wrapper


@init_robokudo_interface
def send_query(obj_type: Optional[str] = None, region: Optional[str] = None,
               attributes: Optional[List[str]] = None) -> Any:
    """Generic function to send a query to RoboKudo."""

    global client
    goal = QueryGoal()

    if obj_type:
        goal.obj.type = obj_type
    if region:
        goal.obj.location = region
    if attributes:
        goal.obj.attribute = attributes

    query_result = None

    def done_callback(state, result):
        nonlocal query_result
        query_result = result
        loginfo("Query completed with state: %s" % state)

    def active_callback():
        loginfo("Goal is now being processed by the action server")

    def feedback_callback(feedback):
        loginfo("Received feedback: %s" % feedback)

    client.send_goal(goal, done_cb=done_callback, active_cb=active_callback, feedback_cb=feedback_callback)
    loginfo("Goal has been sent to the action server")

    client.wait_for_result()
    loginfo("Waiting for result from the action server")
    return query_result


@init_robokudo_interface
def query_all_objects() -> dict:
    """Query RoboKudo for all objects."""
    result = send_query()

    return result


@init_robokudo_interface
def query_object(obj_desc: ObjectDesignatorDescription) -> dict:
    """Query RoboKudo for an object that fits the description."""
    goal = QueryGoal()
    goal.obj.type = str(obj_desc.types[0])

    result = send_query(obj_type=goal.obj.type)

    return result


@init_robokudo_interface
def query_human() -> 'PointStamped':
    """Query RoboKudo for human detection and return the detected human's pose."""
    result = send_query(obj_type='human')
    if result:
        return result  # Assuming result is of type PointStamped or similar.
    return None


@init_robokudo_interface
def stop_query():
    """Stop any ongoing query to RoboKudo."""
    global client
    client.cancel_all_goals()
    loginfo("Cancelled current RoboKudo query goal")


@init_robokudo_interface
def query_specific_region(region: str) -> Any:
    """Query RoboKudo to scan a specific region."""
    return send_query(region=region)


@init_robokudo_interface
def query_human_attributes() -> Any:
    """Query RoboKudo for human attributes like brightness of clothes, headgear, and gender."""
    return send_query(obj_type='human', attributes=["attributes"])


@init_robokudo_interface
def query_waving_human() -> PoseStamped:
    """Query RoboKudo for detecting a waving human."""
    result = send_query(obj_type='human')
    if result and result.res:
        try:
            pose = PoseStamped.from_pose_stamped(result.res[0].pose[0])
            return pose
        except IndexError:
            pass
    return None



from . import node
from .logging import logwarn
import rclpy
from rcl_interfaces.srv import GetParameters
from ament_index_python.packages import get_package_share_directory

def get_node_names(namespace=None):
    """
    Get the names of all nodes in the ROS system.

    :param namespace: The namespace to search for nodes.
    :return: A list of node names.
    """
    return node.get_node_names()

def create_ros_pack(ros_paths=None):
    logwarn("create_ros_pack is not implemented in ROS 2. Please use get_ros_package_path.")
    return None

def get_ros_package_path(package_name):
    return get_package_share_directory(package_name=package_name)

def get_parameter(name):
    try:
        node_names_and_namespaces = node.get_node_names_and_namespaces()
        for node_name, namespace in node_names_and_namespaces:
            client = node.create_client(GetParameters, f'{namespace}/{node_name}/get_parameters')
            request = GetParameters.Request()
            request.names = [name]

            future = client.call_async(request)
            rclpy.spin_until_future_complete(node, future)
            client.destroy()
            if future.result() is not None and future.result().values:
                value = future.result().values[0].value
                return value
    except Exception as e:
        print(f"Failed to get parameter '{name}': {e}")

def wait_for_message(topic_name, msg_type):
    last_message = None
    def sub_callback(msg):
        global last_message
        last_message = msg
        sub.destroy()

    sub = node.create_subscription(msg_type, topic_name, sub_callback())
    return last_message

def is_master_online():
    return node is not None

def sleep(duration):
    """
    Sleep for a given duration.

    :param duration: The duration to sleep in seconds.
    """
    rate = node.create_rate(1 / duration)
    rate.sleep()

def get_time():
    return node.get_time()

def create_timer(duration, callback, oneshot=False):
    timer = node.create_timer(duration, callback, autostart=True)
    return timer

class ResourceNotFound(Exception):
    def __init__(self, *args, **kwargs):
        """Create a new plan failure."""
        Exception.__init__(self, *args, **kwargs)

class ServiceException(Exception):
    def __init__(self, *args, **kwargs):
        """Create a new plan failure."""
        Exception.__init__(self, *args, **kwargs)




import re
import xml.etree.ElementTree as ET
import os
from typing import List, Any

import pycrap
from pycrap.ontologies import FixedJoint, PlanarJoint, RevoluteJoint, FloatingJoint, HingeJoint, ContinuousJoint, \
    PrismaticJoint

def remove_digits(string):
    return re.sub(r'\d+', '', string)



def parse_furniture(link: str):
    """
    Matching the link names from the parsed description with the given classes in the Ontology.
    This procedure should be replaced with Ontology Tinder.
    """
    #furniture = ["Door", "Fridge", "Handle", "Sink", "Drawer", "Table", "Leg"]
    matched_furniture = []

    for l in link.split('_'):
        for c in pycrap.ontologies.base.ontology.classes():
            if c == pycrap.ontologies.Base:
                continue
            clean_l = remove_digits(l).capitalize()  # Remove digits and capitalize
            # print(f"clsname: {c[0]}, link: {clean_l}")
            splitted = str(c).split(".")

            # TODO: I dont get why this can happen. All ontology classes should be prefixed with some module name.
            if len(splitted) == 1:
                return None
            if clean_l == splitted[1]:
                matched_furniture.append(c)
    if matched_furniture:
        return matched_furniture[len(matched_furniture)-1]

def parse_joint_types(joint):

    """
    Parsing the Joint types coming from the joint description, to be added as type of the corresponding classes
    of PyCRAP.
    """
    joint_types = [FixedJoint, PlanarJoint, RevoluteJoint, FloatingJoint, HingeJoint, ContinuousJoint, PrismaticJoint]
    for j in joint_types:
        if str(joint).lower() in str(j).lower():
            return j


from typing_extensions import List, Union, Optional
from numpy.linalg import norm
from numpy import array
from geometry_msgs.msg import Vector3

from owlready2 import Thing

from ...action_designator import PickUpAction, PickUpAction
from ....local_transformer import LocalTransformer
from ....datastructures.world import World
from ....datastructures.pose import PoseStamped, TransformStamped
from ....datastructures.enums import Arms, Grasp
from ....robot_description import RobotDescription, KinematicChainDescription
from ....designator import ObjectDesignatorDescription
from ....ros import loginfo


class DualArmPickupAction(PickUpAction):
    """
    Specialization version of the PickUpAction designator which uses heuristics to solve for a dual pickup solution.
    """

    def __init__(self,
                 object_designator_description: Union[ObjectDesignatorDescription, ObjectDesignatorDescription.Object],
                 grasps: List[Grasp], resolver=None):
        """
        Specialized version of the PickUpAction designator which uses heuristics to solve for a dual pickup problem. The
        designator will choose the arm which is closest to the object that is to be picked up.

        :param object_designator_description: List of object designator which should be picked up
        :param grasps: List of possible grasps which should be used for the pickup
        :param resolver: Optional specialized_designators that returns a performable designator with elements from the
                         lists of possible parameter
        """
        super().__init__(object_designator_description,
                         arms=[Arms.LEFT, Arms.RIGHT],
                         grasps=grasps,
                         resolver=resolver)

        self.object_designator_description: Union[
            ObjectDesignatorDescription, ObjectDesignatorDescription.Object] = object_designator_description

        left_gripper = RobotDescription.current_robot_description.get_arm_chain(Arms.LEFT)
        right_gripper = RobotDescription.current_robot_description.get_arm_chain(Arms.RIGHT)
        self.gripper_list: List[KinematicChainDescription] = [left_gripper, right_gripper]


    def ground(self) -> PickUpAction:
        if isinstance(self.object_designator_description, ObjectDesignatorDescription.Object):
            obj_desig = self.object_designator_description
        else:
            obj_desig = self.object_designator_description.resolve()

        loginfo("Calculating closest gripper to object {}".format(obj_desig.name))

        local_transformer = LocalTransformer()

        object_pose: PoseStamped = obj_desig.world_object.pose
        distances = []
        # Iterate over possible grippers
        for gripper in self.gripper_list:
            # Object pose in gripper frame
            gripper_frame = World.robot.get_link_tf_frame(gripper.get_tool_frame())

            object_T_gripper: PoseStamped = local_transformer.transform_pose(object_pose, gripper_frame)
            object_V_gripper: Vector3 = object_T_gripper.pose.position  # translation vector
            distance = norm(array([object_V_gripper.x, object_V_gripper.y, object_V_gripper.z]))
            loginfo(f"Distance between {gripper} and {obj_desig.name}: {distance}")
            distances.append(distance)

        min_index = distances.index(min(distances))
        winner = self.gripper_list[min_index]
        loginfo(f"Winner is {winner.arm_type.name} with distance {min(distances):.2f}")

        return PickUpAction(object_designator=obj_desig, arm=winner.arm_type, grasp=self.grasps[0])



from __future__ import annotations

import numpy as np
import pinocchio
from numpy.linalg import norm, solve
from typing_extensions import List, Tuple, Dict

from ..failures import IKError
from ..world_concepts.world_object import Object
from ..datastructures.pose import PoseStamped
from ..config.ik_conf import PinocchioConfig


def create_joint_configuration(robot: Object, model) -> np.ndarray[float]:
    """
    Create a joint configuration vector (q) from the current joint positions of the robot.

    :param robot: The robot object.
    :param model: The Pinocchio model.
    :return: The joint configuration vector.
    """
    configuration = []
    for name, joint in zip(model.names, model.joints):
        if name == "universe":
            continue
        # These are continuous joints for all dimensions which are represented by the sin and cos values
        if joint.shortname() in ["JointModelRUBX", "JointModelRUBY", "JointModelRUBZ", "JointModelRevoluteUnboundedUnaligned"]:
            configuration.append(np.cos(robot.joints[name].position))
            configuration.append(np.sin(robot.joints[name].position))
        else:
            # Every other joint of a robot has only one degree of freedom hence they will be directly added to the
            # configuration vector
            configuration.append(robot.joints[name].position)
    return np.array(configuration)


def compute_ik(target_link: str, target_pose: PoseStamped, robot: Object) -> Dict[str, float]:
    """
    Compute the inverse kinematics for a given target link and pose.

    :param target_link: The target link.
    :param target_pose: The target pose.
    :param robot: The robot object.
    :return: The joint configuration as a dictionary with joint names and joint values.
    """
    # Kinematic model of the robot, created form the URDF file
    model = pinocchio.buildModelFromUrdf(robot.path)
    # Mutable data of the robot, includes joint states and position of joints and links
    data = model.createData()

    JOINT_ID = model.frames[model.getFrameId(target_link)].parent
    # Object to destination transformation
    oMdes = pinocchio.XYZQUATToSE3(np.array(target_pose.position.to_list() + target_pose.orientation.to_list()))

    # Initial joint configuration
    # The configuration vector for all joints of the robot, created from the current joint states of the robot
    q = create_joint_configuration(robot, model)
    eps = PinocchioConfig.error_threshold
    IT_MAX = PinocchioConfig.max_iterations
    DT = PinocchioConfig.time_step
    damp = PinocchioConfig.dampening_factor

    q, success = inverse_kinematics_logarithmic(model, q, data, JOINT_ID, oMdes, eps, IT_MAX, DT, damp)

    if success:
        return parse_configuration_vector_to_joint_positions(q, model)
    else:
        raise IKError(pinocchio.SE3ToXYZQUAT(oMdes), robot.tf_frame, target_link)


def inverse_kinematics_logarithmic(model, configuration, data, target_joint_id, target_transformation, eps=1e-4,
                                   max_iter=1000, dt=1e-1, damp=1e-12) -> Tuple[np.ndarray[float], bool]:
    """
    Compute the inverse kinematics for a given target transformation. Using a logarithmic error metric.

    :param model: The Pinocchio model.
    :param configuration: The initial joint configuration.
    :param data: The Pinocchio data.
    :param target_joint_id: The target joint ID.
    :param target_transformation: The target transformation.
    :param eps: The error threshold.
    :param max_iter: The maximum number of iterations.
    :param dt: The time step.
    :param damp: The damping factor.
    :return: The final joint configuration and a boolean indicating if the computation was successful.
    """
    q = configuration
    success = False
    for i in range(max_iter):
        # Calculate the current state of all joints for the current iteration
        pinocchio.forwardKinematics(model, data, q)
        # Transformation between the target pose and the current pose of the joint which should reach the target pose
        iMd = data.oMi[target_joint_id].actInv(target_transformation)
        # Error between the current pose and the target pose, this should be minimized
        err = pinocchio.log(iMd).vector  # in joint frame
        if norm(err) < eps:
            success = True
            break
        J = pinocchio.computeJointJacobian(model, data, q, target_joint_id)  # in joint frame
        J = -np.dot(pinocchio.Jlog6(iMd.inverse()), J)
        v = -J.T.dot(solve(J.dot(J.T) + damp * np.eye(6), err))
        q = pinocchio.integrate(model, q, v * dt)
        # Clip all joint values to their limits since the algorithm does not respect joint limits
        q = clip_joints_to_limits(q, model.lowerPositionLimit, model.upperPositionLimit)

    return q, success


def inverse_kinematics_translation(model, configuration, data, target_joint_id, target_transformation, eps=1e-4,
                                   max_iter=1000, dt=1e-1, damp=1e-12) -> Tuple[np.ndarray[float], bool]:
    """
    Compute the inverse kinematics for a given target transformation. Using the distance of the translation as error metric.

    :param model: The Pinocchio model.
    :param configuration: The initial joint configuration.
    :param data: The Pinocchio data.
    :param target_joint_id: The target joint ID.
    :param target_transformation: The target transformation.
    :param eps: The error threshold.
    :param max_iter: The maximum number of iterations.
    :param dt: The time step.
    :param damp: The damping factor.
    :return: The final joint configuration and a boolean indicating if the computation was successful.
    """
    q = configuration
    success = False
    for i in range(max_iter):
        # Calculate the current state of all joints for the current iteration
        pinocchio.forwardKinematics(model, data, q)
        # Transformation between the target pose and the current pose of the joint which should reach the target pose
        iMd = data.oMi[target_joint_id].actInv(target_transformation)
        # Error metric that should be minimized, in this case this is the translation (x, y, z) without the orientation
        err = iMd.translation
        if norm(err) < eps:
            success = True
            break
        J = pinocchio.computeJointJacobian(model, data, q, target_joint_id)  # in joint frame
        J = -J[:3, :]  # linear part of the Jacobian
        v = -J.T.dot(solve(J.dot(J.T) + damp * np.eye(3), err))
        q = pinocchio.integrate(model, q, v * dt)
        # Clip all joint values to their limits since the algorithm does not respect joint limits
        q = clip_joints_to_limits(q, model.lowerPositionLimit, model.upperPositionLimit)

    return q, success


def parse_configuration_vector_to_joint_positions(configuration: np.ndarray[float], model) -> Dict[str, float]:
    """
    Takes the configuration vector from pinocchio and the robot model and returns a dictionary with joint names and
    joint values.

    :param configuration: The configuration vector.
    :param model: The robot model.
    :return: The joint configuration as a dictionary with joint names and joint values.
    """
    result = {}
    for joint in model.joints:
        if joint.idx_q == -1:
            continue
        if joint.shortname() in ["JointModelRUBX", "JointModelRUBY", "JointModelRUBZ", "JointModelRevoluteUnboundedUnaligned"]:
            # Continuous Joints are represented as sin(theta) and cos(theta) of the joint value, so we use arctan2 to
            # get the joint value
            joint_value = np.arctan2(configuration[joint.idx_q + 1], configuration[joint.idx_q])
        else:
            joint_value = configuration[joint.idx_q]

        result[model.names[joint.id]] = joint_value
    return result


def clip_joints_to_limits(joint_positions: List[float], lower_limits: List[float], upper_limits: List[float]) -> \
np.ndarray[float]:
    """
    Clip the joint positions to the joint limits.

    :param joint_positions: The joint positions to clip.
    :param lower_limits: A List of lower joint limits.
    :param upper_limits: A List of upper joint limits.
    :return: The clipped joint positions.
    """
    return np.clip(joint_positions, lower_limits, upper_limits)



from dataclasses import dataclass
from datetime import datetime
from typing import Type

from ormatic.ormatic import ORMaticExplicitMapping
from ormatic.utils import classproperty
from typing_extensions import Optional

from pycrap.ontologies import PhysicalObject
from .casts import StringType
from ..datastructures.dataclasses import FrozenObject as FrozenObjectIm
from ..datastructures.enums import Arms, DetectionTechnique, DetectionState, Grasp, TaskStatus
from ..datastructures.grasp import GraspDescription
from ..datastructures.pose import Pose, PoseStamped, Vector3
from ..designators.action_designator import MoveTorsoAction, SetGripperAction, ReleaseAction as ReleaseActionDesignator, \
    GripAction as GripActionDesignator, ParkArmsAction, PickUpAction as PickUpActionDesignator, \
    PlaceAction as PlaceActionDesignator, NavigateAction, TransportAction as TransportActionDesignator, LookAtAction, \
    DetectAction as DetectActionDesignator, OpenAction, CloseAction, GraspingAction as GraspingActionDesignator, \
    FaceAtAction, ActionDescription, ReachToPickUpAction as ReachToPickUpActionDesignator, SearchAction as SearchActionDesignator
from pycram.plan import ActionNode, MotionNode, ResolvedActionNode


# ----------------------------------------------------------------------------------------------------------------------
#            Map all Designators, that are not self-mapping, here.
#            By default all classes are self-mapping, so you only need to add the ones where not every attribute is
#            supposed to be mapped or where an attribute is from a type, which is not mapped itself.
#            Specify the columns(attributes) that are supposed to be tracked in the database.
#            One attribute equals one column. Please refer to the ORMatic documentation for more information.
# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class TaskTreeNode:
    start_time: datetime
    end_time: Optional[datetime]
    status: TaskStatus
    reason: Optional[str]
    action: Optional[ActionDescription] = None
    parent: Optional["TaskTreeNode"] = None


@dataclass
class ORMActionNode(ORMaticExplicitMapping):

    @classproperty
    def explicit_mapping(cls) -> Type:
        return ActionNode


@dataclass
class ORMMotionNode(ORMaticExplicitMapping):

    @classproperty
    def explicit_mapping(cls) -> Type:
        return MotionNode


@dataclass
class ORMResolvedActionNode(ORMaticExplicitMapping):
    designator_ref: ActionDescription

    @classproperty
    def explicit_mapping(cls) -> Type:
        return ResolvedActionNode


@dataclass
class FrozenObject(ORMaticExplicitMapping):
    name: str
    concept: PhysicalObject
    pose: Optional[PoseStamped]

    @classproperty
    def explicit_mapping(cls):
        return FrozenObjectIm


@dataclass
class ReleaseAction(ActionDescription, ORMaticExplicitMapping):
    gripper: Arms
    object_at_execution: Optional[FrozenObject]

    @classproperty
    def explicit_mapping(cls):
        return ReleaseActionDesignator


@dataclass
class GripAction(ActionDescription, ORMaticExplicitMapping):
    gripper: Arms
    effort: float
    object_at_execution: Optional[FrozenObject]

    @classproperty
    def explicit_mapping(cls):
        return GripActionDesignator


@dataclass
class ReachToPickUpAction(ActionDescription, ORMaticExplicitMapping):
    arm: Arms
    grasp_description: GraspDescription
    object_at_execution: Optional[FrozenObject]

    @classproperty
    def explicit_mapping(cls):
        return ReachToPickUpActionDesignator


@dataclass
class PickUpAction(ActionDescription, ORMaticExplicitMapping):
    arm: Arms
    object_at_execution: Optional[FrozenObject]

    @classproperty
    def explicit_mapping(cls):
        return PickUpActionDesignator


@dataclass
class PlaceAction(ActionDescription, ORMaticExplicitMapping):
    arm: Arms
    target_location: PoseStamped
    object_at_execution: Optional[FrozenObject]

    @classproperty
    def explicit_mapping(cls):
        return PlaceActionDesignator


@dataclass
class TransportAction(ActionDescription, ORMaticExplicitMapping):
    arm: Arms
    target_location: PoseStamped
    object_at_execution: Optional[FrozenObject]

    @classproperty
    def explicit_mapping(cls):
        return TransportActionDesignator


@dataclass
class DetectAction(ActionDescription, ORMaticExplicitMapping):
    technique: DetectionTechnique
    state: Optional[DetectionState]
    region: Optional[str]
    object_at_execution: Optional[FrozenObject]

    @classproperty
    def explicit_mapping(cls):
        return DetectActionDesignator


@dataclass
class GraspingAction(ActionDescription, ORMaticExplicitMapping):
    arm: Arms
    object_at_execution: FrozenObject

    @classproperty
    def explicit_mapping(cls):
        return GraspingActionDesignator


@dataclass
class MoveAndPickUpAction(ActionDescription, ORMaticExplicitMapping):
    standing_position: Pose
    object_designator: FrozenObject
    arm: Arms
    grasp: Grasp
    keep_joint_states: bool

    # @classproperty
    # def explicit_mapping(cls):
    #     return

# specify custom type mappings
type_mappings ={
    PhysicalObject: StringType(),
}


# List of all classes that are self-mapped. ADD NEW DESIGNATORS HERE!
self_mapped_classes = [TaskTreeNode, ActionDescription, MoveTorsoAction, SetGripperAction, ParkArmsAction,
                       NavigateAction, LookAtAction, OpenAction, CloseAction, FaceAtAction, SearchActionDesignator]

# List of all classes that are explicitly mapped above. ADD NEW DESIGNATORS HERE!
explicitly_mapped_classes = [DetectAction, GraspingAction, ReleaseAction, GripAction, FrozenObject,
                             ReachToPickUpAction, PickUpAction, PlaceAction, TransportAction, ORMActionNode,
                             ORMMotionNode, ORMResolvedActionNode]



from ..language import LanguageMixin
from ..world_concepts.world_object import Object, World
from ..datastructures.enums import ObjectType
from ..datastructures.pose import PoseStamped
import gymnasium as gym
import pycrap

class PyCRAMGym(gym.Env):
    """
    First small implementation of the PyCRAMGYM, this class can either be used to store necessary information for an experiment or let
    a robot explore an environment, with a set of actions. The exploration will be done via reinforcement learning.
    """

    world: str # should be World currently not possible because spawning multiple bullet worlds leads to problems
    """
    The world should be an real World Object, but sadly this is not possible, because if we close and open multiple 
    bulletworlds in quick succession, then there is a problem with objects not being loaded properly. Thus we decided to
    let the user create the world they want to use and just load the environment urdf of the file they give to this class
    """
    robot: str # name to search for robot
    """
    The robot should be a real Robot Object, but in order to stay in line with the world it is just the name of the 
    urdf that will be loaded as a robot.
    """
    plan: LanguageMixin # plan language object that can performed
    """
    Plan langauge object that can get executed. 
    """

    def __init__(self, inWorld: str, inRobot: str, inPlan: LanguageMixin):
        super().__init__()
        if World.current_world is not None:
            if inRobot.endswith('.urdf'):
                self.robot = Object(inRobot.split('.urdf')[0], pycrap.Robot, inRobot, pose=PoseStamped.from_list([1, 2, 0]))
            else:
                self.robot = Object(inRobot, pycrap.Robot, inRobot +".urdf", pose=PoseStamped.from_list([1, 2, 0]))
            self.world = Object(inWorld, pycrap.Apartment, inWorld + ".urdf")
            self.plan = inPlan

    def step(self, action: gym.core.ActType):
        """
        Ticks the world, in our case this means we pick one possible action and perform it.

        :ivar action: The action to take.
        """
        return None

    def reset(self):
        """
        Returns the environment to the normal state.
        """
        self.world.reset()

    def close(self):
        """
        Resets the world to a completely empty state. (Should close world at some point)
        """
        if World.current_world is not None and self.world is not None:
            World.current_world.remove_object(self.world)
            World.current_world.remove_object(self.robot)



# Funky MonkeyPatching for ROS2 compatability
import builtin_interfaces.msg
def to_sec(self):
    """
    Returns the time in seconds from a builtin_interfaces.msg.Time message.

    :return: The time in seconds.
    """
    return self.sec

builtin_interfaces.msg.Time.to_sec = to_sec

import rclpy
import threading
from rclpy.node import Node

rclpy.init()
node = Node('pycram')
threading.Thread(target=rclpy.spin, args=(node,), daemon=True).start()

from .data_types import *
from .ros_tools import  *
from .logging import *
from .action_lib import *
from .service import *
from .publisher import *
from .subscriber import *
from .viz_marker_publisher import *


import unittest


class TestGiskardInterface(unittest.TestCase):
    def test_something(self):

        self.assertEqual(True, True)  # add assertion here


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import timedelta

from pycram.worlds.bullet_world import BulletWorld, Object
from pycram.designator import ActionDescription
from pycram.designators.action_designator import ParkArmsAction, ParkArmsActionDescription
from pycram.datastructures.enums import ObjectType, Arms, WorldMode
from pycram.failure_handling import Retry
from pycram.failures import PlanFailure
from pycram.process_module import ProcessModule, simulated_robot
from pycram.robot_description import RobotDescription
from pycram.object_descriptors.urdf import ObjectDescription
from pycrap.ontologies import Robot

extension = ObjectDescription.get_file_extension()


# start ik_and_description.launch
class DummyActionDesignator(ActionDescription):
    class Action(ActionDescription):
        def perform(self):
            raise PlanFailure("Dummy action failed")

    def __iter__(self):
        for _ in range(100):
            yield self.Action()


class FailureHandlingTest(unittest.TestCase):
    world: BulletWorld

    @classmethod
    def setUpClass(cls):
        cls.world = BulletWorld(WorldMode.DIRECT)
        cls.robot = Object(RobotDescription.current_robot_description.name, Robot,
                           RobotDescription.current_robot_description.name + extension)
        ProcessModule.execution_delay = timedelta(seconds=0.5)

    def setUp(self):
        self.world.reset_world()

    def test_retry_with_success(self):
        with simulated_robot:
            Retry(ParkArmsActionDescription([Arms.LEFT]), max_tries=5).perform()

    def test_retry_with_failure(self):
        with simulated_robot:
            with self.assertRaises(PlanFailure):
                Retry(DummyActionDesignator(), max_tries=5).perform()

    def tearDown(self):
        self.world.reset_world()

    @classmethod
    def tearDownClass(cls):
        cls.world.exit()


if __name__ == '__main__':
    unittest.main()



from ..tf_transformations import quaternion_from_euler

from ....datastructures.enums import StaticJointState
from ....datastructures.pose import PoseStamped
from ....designators.location_designator import CostmapLocation
from ....external_interfaces.giskard import projection_cartesian_goal_with_approach, projection_joint_goal
from ....robot_description import RobotDescription
from ....datastructures.world import UseProspectionWorld, World
from ....local_transformer import LocalTransformer
from ....costmaps import OccupancyCostmap, GaussianCostmap
from ....pose_generator_and_validator import PoseGenerator


class GiskardLocation(CostmapLocation):
    """'
    Specialization version of the CostmapLocation which uses Giskard to solve for a full-body IK solution. This
    designator is especially useful for robots which lack a degree of freedom and therefore need to use the base to
    manipulate the environment effectively.
    """

    def __iter__(self) -> CostmapLocation.Location:
        """
        Uses Giskard to perform full body ik solving to get the pose of a robot at which it is able to reach a certain point.

        :yield: An instance of CostmapLocation.Location with a pose from which the robot can reach the target
        """
        if self.visible_for:
            raise ValueError("GiskardLocation does not support the visible_for parameter")
        local_transformer = LocalTransformer()
        target_map = local_transformer.transform_pose(self.target, "map")

        manipulator_descs = RobotDescription.current_robot_description.get_manipulator_chains()

        near_costmap = (OccupancyCostmap(0.35, False, 200, 0.02, target_map)
                        + GaussianCostmap(200, 15, 0.02, target_map))
        for maybe_pose in PoseGenerator(near_costmap, 200):
            for chain in manipulator_descs:
                projection_joint_goal(chain.get_static_joint_states(StaticJointState.Park), allow_collisions=True)

                trajectory = projection_cartesian_goal_with_approach(maybe_pose, target_map, chain.tool_frame,
                                                                     "map", RobotDescription.current_robot_description.base_link)
                last_point_positions = trajectory.trajectory.points[-1].positions
                last_point_names = trajectory.trajectory.joint_names
                last_joint_states = dict(zip(last_point_names, last_point_positions))
                orientation = list(
                    quaternion_from_euler(0, 0, last_joint_states["brumbrum_yaw"], axes="sxyz"))
                pose = PoseStamped.from_list([last_joint_states["brumbrum_x"], last_joint_states["brumbrum_y"], 0], orientation)

                robot_joint_states = {}
                for joint_name, state in last_joint_states.items():
                    if joint_name in World.robot.joints.keys():
                        robot_joint_states[joint_name] = state

                prospection_robot = World.current_world.get_prospection_object_for_object(World.robot)

                with UseProspectionWorld():
                    prospection_robot.set_multiple_joint_positions(robot_joint_states)
                    prospection_robot.set_pose(pose)
                    gripper_pose = prospection_robot.get_link_pose(chain.get_tool_frame())

                    if gripper_pose.position.euclidean_distance(target_map.position) <= 0.02:
                        yield CostmapLocation.Location(pose, [chain.arm_type])





import atexit
import time
import threading
from geometry_msgs.msg import WrenchStamped
from std_msgs.msg import Header

from ..datastructures.enums import FilterConfig
from ..datastructures.world import World
from ..failures import SensorMonitoringCondition
from ..filter import Butterworth
from ..ros import  Time
from ..ros import  create_publisher, logdebug, loginfo_once, logerr, create_subscriber


class ForceTorqueSensorSimulated:
    """
    Simulated force-torque sensor for a joint with a given name.
    Reads simulated forces and torques at that joint from world and publishes geometry_msgs/Wrench messages
    to the given topic.
    """

    def __init__(self, joint_name, fts_topic="/pycram/fts", interval=0.1):
        """
        The given joint_name has to be part of :py:attr:`~pycram.world.World.robot` otherwise a
        RuntimeError will be raised.

        :param joint_name: Name of the joint for which force-torque should be simulated
        :param fts_topic: Name of the ROS topic to which should be published
        :param interval: Interval at which the messages should be published, in seconds
        """
        self.world = World.current_world
        self.fts_joint_idx = None
        self.joint_name = joint_name
        if joint_name in self.world.robot.joint_name_to_id.keys():
            self.fts_joint_idx = self.world.robot.joint_name_to_id[joint_name]
        else:
            raise RuntimeError(f"Could not register ForceTorqueSensor: Joint {joint_name}"
                               f" does not exist in robot object")
        self.world.enable_joint_force_torque_sensor(self.world.robot, self.fts_joint_idx)

        self.fts_pub = create_publisher(fts_topic, WrenchStamped, queue_size=10)
        self.interval = interval
        self.kill_event = threading.Event()

        self.thread = threading.Thread(target=self._publish)
        self.thread.start()

        atexit.register(self._stop_publishing)

    def _publish(self) -> None:
        """
        Continuously publishes the force-torque values for the simulated joint. Values are published as long as the
        kill_event is not set.
        """
        seq = 0
        while not self.kill_event.is_set():
            joint_ft = self.world.get_joint_reaction_force_torque(self.world.robot, self.fts_joint_idx)
            h = Header()
            h.stamp = Time().now()
            h.frame_id = self.joint_name

            wrench_msg = WrenchStamped()
            wrench_msg.header = h
            wrench_msg.wrench.force.x = joint_ft[0]
            wrench_msg.wrench.force.y = joint_ft[1]
            wrench_msg.wrench.force.z = joint_ft[2]

            wrench_msg.wrench.torque.x = joint_ft[3]
            wrench_msg.wrench.torque.y = joint_ft[4]
            wrench_msg.wrench.torque.z = joint_ft[5]

            self.fts_pub.publish(wrench_msg)
            seq += 1
            time.sleep(self.interval)

    def _stop_publishing(self) -> None:
        """
        Sets the kill_event and therefore terminates the Thread publishing the force-torque values as well as join the
        threads.
        """
        self.kill_event.set()
        self.thread.join()


class ForceTorqueSensor:
    """
    Monitor a force-torque sensor of a supported robot and save relevant data.

    Apply a specified filter and save this data as well.
    Default filter is the low pass filter 'Butterworth'

    Can also calculate the derivative of (un-)filtered data

    :param robot_name: Name of the robot
    :param filter_config: Desired filter (default: Butterworth)
    :param filter_order: Order of the filter. Declares the number of elements that delay the sampling
    :param custom_topic: Declare a custom topic if the default topics do not fit
    """

    filtered = 'filtered'
    unfiltered = 'unfiltered'

    def __init__(self, robot_name, filter_config=FilterConfig.butterworth, filter_order=4, custom_topic=None,
                 debug=False):
        self.robot_name = robot_name
        self.filter_config = filter_config
        self.filter = self._get_filter(order=filter_order)
        self.debug = debug

        self.wrench_topic_name = custom_topic
        self.force_torque_subscriber = None
        self.init_data = True

        self.whole_data = None
        self.prev_values = None

        self.order = filter_order

        self._setup()

    def _setup(self):
        self._get_robot_parameters()
        self.subscribe()

    def _get_robot_parameters(self):
        if self.wrench_topic_name is not None:
            return

        if self.robot_name == 'hsrb':
            self.wrench_topic_name = '/hsrb/wrist_wrench/compensated'

        elif self.robot_name == 'iai_donbot':
            self.wrench_topic_name = '/kms40_driver/wrench'

        elif self.robot_name == 'pr2':
            raw_data = "/ft/l_gripper_motor_zeroed"
            raw_data_10_sample_moving_average = "/ft/l_gripper_motor_zeroed_avg"
            derivative_data = "/ft/l_gripper_motor_zeroed_derivative"
            derivative_data_10_sample_moving_average = "/ft/l_gripper_motor_zeroed_derivative_avg"

            self.wrench_topic_name = raw_data
        else:
            logerr(f'{self.robot_name} is not supported')

    def _get_rospy_data(self, data_compensated: WrenchStamped):
        if self.init_data:
            self.init_data = False
            self.prev_values = [data_compensated] * (self.order + 1)
            self.whole_data = {self.unfiltered: [data_compensated],
                               self.filtered: [data_compensated]}

        filtered_data = self._filter_data(data_compensated)

        self.whole_data[self.unfiltered].append(data_compensated)
        self.whole_data[self.filtered].append(filtered_data)

        self.prev_values.append(data_compensated)
        self.prev_values.pop(0)

        if self.debug:
            logdebug(
                f'x: {data_compensated.wrench.force.x}, '
                f'y: {data_compensated.wrench.force.y}, '
                f'z: {data_compensated.wrench.force.z}')

    def _get_filter(self, order=4, cutoff=10, fs=60):
        if self.filter_config == FilterConfig.butterworth:
            return Butterworth(order=order, cutoff=cutoff, fs=fs)

    def _filter_data(self, current_wrench_data: WrenchStamped) -> WrenchStamped:
        filtered_data = WrenchStamped()
        filtered_data.header = current_wrench_data.header
        for attr in ['x', 'y', 'z']:
            force_values = [getattr(val.wrench.force, attr) for val in self.prev_values] + [
                getattr(current_wrench_data.wrench.force, attr)]
            torque_values = [getattr(val.wrench.torque, attr) for val in self.prev_values] + [
                getattr(current_wrench_data.wrench.torque, attr)]

            filtered_force = self.filter.filter(force_values)[-1]
            filtered_torque = self.filter.filter(torque_values)[-1]

            setattr(filtered_data.wrench.force, attr, filtered_force)
            setattr(filtered_data.wrench.torque, attr, filtered_torque)

        return filtered_data

    def subscribe(self):
        """
        Subscribe to the specified wrench topic.

        This will automatically be called on setup.
        Only use this if you already unsubscribed before.
        """
        self.force_torque_subscriber = create_subscriber(self.wrench_topic_name,
                                                            WrenchStamped,
                                                            self._get_rospy_data)

    def unsubscribe(self):
        """
        Unsubscribe from the specified topic
        """
        self.force_torque_subscriber.unregister()

    def get_last_value(self, is_filtered=True) -> WrenchStamped:
        """
        Get the most current data values.

        :param is_filtered: Decides about using filtered or raw data

        :return: A list containing the most current values (newest are first)
        """
        status = self.filtered if is_filtered else self.unfiltered
        return self.whole_data[status][-1]

    def get_derivative(self, is_filtered=True) -> WrenchStamped:
        """
        Calculate the derivative of current data.

        :param is_filtered: Decides about using filtered or raw data.
        :return: The derivative as a WrenchStamped object. Returns a zeroed derivative if only one data point exists.
        """
        status = self.filtered if is_filtered else self.unfiltered

        if len(self.whole_data[status]) < 2:
            return WrenchStamped()

        before: WrenchStamped = self.whole_data[status][-2]
        after: WrenchStamped = self.whole_data[status][-1]
        derivative = WrenchStamped()

        dt = (after.header.stamp - before.header.stamp).to_sec()
        if dt == 0:
            return WrenchStamped()

        derivative.wrench.force.x = float((after.wrench.force.x - before.wrench.force.x) / dt)
        derivative.wrench.force.y = float((after.wrench.force.y - before.wrench.force.y) / dt)
        derivative.wrench.force.z = float((after.wrench.force.z - before.wrench.force.z) / dt)
        derivative.wrench.torque.x = float((after.wrench.torque.x - before.wrench.torque.x) / dt)
        derivative.wrench.torque.y = float((after.wrench.torque.y - before.wrench.torque.y) / dt)
        derivative.wrench.torque.z = float((after.wrench.torque.z - before.wrench.torque.z) / dt)

        return derivative

    def human_touch_monitoring(self):
        loginfo_once("Now monitoring for human touch")
        if self.robot_name == 'pr2':
            der = self.get_derivative()
            if abs(der.wrench.torque.x) > 3:
                return SensorMonitoringCondition



import unittest

import numpy as np
from random_events.variable import Continuous
#  import plotly.graph_objects as go
from random_events.product_algebra import Event, SimpleEvent
from random_events.interval import *

from pycram.testing import BulletWorldTestCase
from pycram.costmaps import OccupancyCostmap, AlgebraicSemanticCostmap, VisibilityCostmap
from pycram.probabilistic_costmap import ProbabilisticCostmap
from pycram.units import meter, centimeter
from pycram.datastructures.pose import PoseStamped
import plotly.graph_objects as go


class CostmapTestCase(BulletWorldTestCase):

    def test_raytest_bug(self):
        for i in range(30):
            o = OccupancyCostmap(distance_to_obstacle=0.2, from_ros=False, size=200, resolution=0.02,
                                 origin=PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1]))

    def test_attachment_exclusion(self):
        self.kitchen.set_pose(PoseStamped.from_list([50, 50, 0]))
        self.robot.set_pose(PoseStamped.from_list([0, 0, 0]))
        self.milk.set_pose(PoseStamped.from_list([0.5, 0, 1]))
        self.cereal.set_pose(PoseStamped.from_list([50, 50, 0]))
        o = OccupancyCostmap(0.2, from_ros=False, size=200, resolution=0.02,
                             origin=PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1]))
        self.assertEqual(np.sum(o.map[115:135, 90:110]), 0)

        self.robot.attach(self.milk)
        self.assertTrue(np.sum(o.map[80:90, 90:110]) != 0)

    def test_partition_into_rectangles(self):
        ocm = OccupancyCostmap(distance_to_obstacle=0.2, from_ros=False, size=200, resolution=0.02,
                               origin=PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1]))
        rectangles = ocm.partitioning_rectangles()
        ocm.visualize()

        x = Continuous("x")
        y = Continuous("y")

        events = []
        for rectangle in rectangles:

            event = SimpleEvent({x: open(rectangle.x_lower, rectangle.x_upper),
                           y: open(rectangle.y_lower, rectangle.y_upper)})
            events.append(event)

        event = Event(*events)
        # fig = go.Figure(event.plot(), event.plotly_layout())
        # fig.update_xaxes(range=[-2, 2])
        # fig.update_yaxes(range=[-2, 2])
        # fig.show()
        self.assertTrue(event.is_disjoint())

    def test_visualize(self):
        o = OccupancyCostmap(0.2, from_ros=False, size=200, resolution=0.02,
                             origin=PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1]))
        o.visualize()

    def test_merge_costmap(self):
        o = OccupancyCostmap(0.2, from_ros=False, size=200, resolution=0.02,
                             origin=PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1]))
        o2 = OccupancyCostmap(0.2, from_ros=False, size=200, resolution=0.02,
                              origin=PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1]))
        o3 = o + o2
        self.assertTrue(np.all(o.map == o3.map))
        o2.map[100:120, 100:120] = 0
        o3 = o + o2
        self.assertTrue(np.all(o3.map[100:120, 100:120] == 0))
        self.assertTrue(np.all(o3.map[0:100, 0:100] == o.map[0:100, 0:100]))
        o2.map = np.zeros_like(o2.map)
        o3 = o + o2
        self.assertTrue(np.all(o3.map == o2.map))


class SemanticCostmapTestCase(BulletWorldTestCase):

    def test_generate_map(self):
        costmap = AlgebraicSemanticCostmap(self.kitchen, "kitchen_island_surface")
        costmap.valid_area = costmap.valid_area.__deepcopy__() & costmap.left()
        costmap.valid_area = costmap.valid_area.__deepcopy__() & costmap.top()
        costmap.valid_area = costmap.valid_area.__deepcopy__() & costmap.border(0.2)
        self.assertEqual(len(costmap.valid_area.simple_sets), 2)

    def test_as_distribution(self):
        costmap = AlgebraicSemanticCostmap(self.kitchen, "kitchen_island_surface")
        costmap.valid_area = costmap.valid_area.__deepcopy__() & costmap.right() & costmap.bottom() & costmap.border(0.2)
        model = costmap.as_distribution()
        self.assertEqual(len(model.nodes), 7)
        # fig = go.Figure(model.plot(), model.plotly_layout())
        # fig.show()
        # supp = model.support
        # fig = go.Figure(supp.plot(), supp.plotly_layout())
        # fig.show()

    def test_iterate(self):
        costmap = AlgebraicSemanticCostmap(self.kitchen, "kitchen_island_surface")
        costmap.valid_area = costmap.valid_area.__deepcopy__() & costmap.left() & costmap.bottom() & costmap.border(0.2)
        for sample in iter(costmap):
            self.assertIsInstance(sample, PoseStamped)
            self.assertTrue(costmap.valid_area.contains([sample.position.x, sample.position.y]))


@unittest.skip("Wait for PM Upgrade to go live")
class ProbabilisticCostmapTestCase(BulletWorldTestCase):

    origin = PoseStamped.from_list([1.5, 1, 0], [0, 0, 0, 1])

    def setUp(self):
        super().setUp()
        self.costmap = ProbabilisticCostmap(self.origin, size = 200*centimeter)

    def test_setup(self):
        event = self.costmap.create_event_from_map()
        self.assertTrue(event.is_disjoint())

    def test_visualization(self):
        fig = go.Figure(self.costmap.distribution.plot(), self.costmap.distribution.plotly_layout())
        self.costmap.visualize()
        # fig = go.Figure(pcm.distribution.plot(surface=True), pcm.distribution.plotly_layout())
        # fig.show()

    def test_visibility_cm(self):
        costmap = ProbabilisticCostmap(self.origin, size = 200*centimeter,
                                       costmap_type=VisibilityCostmap)
        costmap.visualize()

    def test_merge_cm(self):
        visibility = ProbabilisticCostmap(self.origin, size = 200*centimeter,
                                       costmap_type=VisibilityCostmap)
        occupancy = ProbabilisticCostmap(self.origin, size = 200*centimeter)
        occupancy.distribution, _ = occupancy.distribution.conditional(visibility.create_event_from_map())
        occupancy.visualize()



from datetime import datetime
from functools import lru_cache

from ..designators.action_designator import *
from ..designators.object_designator import *
from ..plan import ActionNode, DesignatorNode, ResolvedActionNode, PlanNode
from ..ros import create_publisher
from ..ros import Time as ROSTime

try:
    from knowrob_designator.msg import DesignatorExecutionFinished, DesignatorExecutionStart, DesignatorInit, \
        DesignatorResolutionFinished, DesignatorResolutionStart, ObjectDesignator
except ImportError:
    loginfo("Could not import knowrob_designator.msg")

desig_execution_start = create_publisher("knowrob/designator_execution_started", DesignatorExecutionStart)
desig_execution_finished = create_publisher("knowrob/designator_execution_finished", DesignatorExecutionFinished)
desig_resolution_start = create_publisher("knowrob/designator_resolving_started", DesignatorResolutionStart)
desig_resolution_finished = create_publisher("knowrob/designator_resolving_finished", DesignatorResolutionFinished)
desig_init = create_publisher("knowrob/designator_init", DesignatorInit)
object_desig = create_publisher("knowrob/object_designator", ObjectDesignator)

@lru_cache(maxsize=None)
def init_object_state():
    """
    Publishes the initial state of the objects in the world to Knowrob. Because of the lru_cache decorator, this function
    will only be called once, and the result will be cached for future calls.
    """
    for obj in World.current_world.objects:
        obj_json = {"anObject": {"type": str(obj.obj_type)}}

        msg = ObjectDesignator()
        msg.json_designator = str(obj_json)
        split_time = str(datetime.now().timestamp()).split(".")
        msg.start = ROSTime(int(split_time[0]), int(split_time[1]))
        object_desig.publish(msg)


def pose_to_json(pose: PoseStamped) -> Dict[str, Union[float, str]]:
    """
    Converts a PoseStamped object to a JSON-like dictionary.

    :return : A dictionary representation of the pose.
    """
    return {"px": pose.position.x, "py": pose.position.y, "pz": pose.position.z, "rx": pose.orientation.x,
            "ry": pose.orientation.y, "rz": pose.orientation.z, "rw": pose.orientation.w, "frame": pose.header.frame_id}


def object_to_json(obj: Union[ObjectDesignator, Object]) -> Dict[str, Dict[str, str]]:
    """
    Converts an object or object designator to a JSON-like dictionary containing its type.

    :param obj: The object or object designator to convert.
    :return: A dictionary representation of the object.
    """
    if isinstance(obj, ObjectDesignatorDescription):
        return {"anObject": {"type": str(obj.types[0])}}
    else:
        return {"anObject": {"type": str(obj.obj_type)}}


def grasp_description_to_json(grasp: GraspDescription) -> Dict[str, Union[float, str]]:
    """
    Converts a GraspDescription object to a JSON-like dictionary.

    :param grasp: The GraspDescription object to convert.
    :return: A dictionary representation of the grasp.
    """
    return {"approach_direction": grasp.approach_direction, "vertical_alignment": grasp.vertical_alignment,
            "rotate_gripper": grasp.rotate_gripper}


def kwargs_to_json(kwargs: Dict) -> dict:
    """
    Converts the keyword arguments of a DesignatorNode to a JSON-like dictionary, calls the appropriate conversion function
    for each value based on its type.

    :param kwargs: The keyword arguments to convert.
    :return: A dictionary representation of the keyword arguments.
    """
    res = {}
    for name, value in kwargs.items():
        if isinstance(value, PoseStamped):
            res[name] = pose_to_json(value)
        elif isinstance(value, Object) or isinstance(value, ObjectDesignatorDescription):
            res[name] = object_to_json(value)
        elif isinstance(value, GraspDescription):
            res[name] = grasp_description_to_json(value)
        else:
            res[name] = value
    return res


def designator_to_json(node: DesignatorNode) -> str:
    """
    Converts a DesignatorNode to a JSON-like string representation uses the kwargs of the node for conversion.

    :param node: The DesignatorNode to convert.
    :return: A JSON-like string representation of the DesignatorNode.
    """
    params = {"type": str(node.action.__name__)}
    params.update(kwargs_to_json(node.kwargs))
    params = {"anAction": params}
    return str(params)

def get_current_ros_time() -> ROSTime:
    """
    Returns the ROS time as a ROSTime object.

    :return: The current ROS time.
    """
    split_time = str(datetime.now().timestamp()).split(".")
    return ROSTime(int(split_time[0]), int(split_time[1]))


def execution_start_callback(node: PlanNode):
    """
    Callback function that is called when the execution of a designator starts. It publishes a message to Knowrob
    containing the designator that is executed.
    """
    msg = DesignatorExecutionStart()
    msg.designator_id = str(hash(node))
    msg.json_designator = designator_to_json(node)
    msg.stamp = get_current_ros_time()

    desig_execution_start.publish(msg)


def execution_finished_callback(node: PlanNode):
    """
    Callback function that is called when the execution of a designator finishes. It publishes a message to Knowrob
    containing the designator that has just finished executing.
    """
    msg = DesignatorExecutionFinished()
    msg.designator_id = str(hash(node))
    msg.json_designator = designator_to_json(node)
    msg.stamp = get_current_ros_time()

    desig_execution_finished.publish(msg)


def resolution_start_callback(node: PlanNode):
    """
    Callback function that is called when the resolution of a designator starts. It publishes a message to Knowrob
    containing the designator that is being resolved. It also sends the initial state of the objects in the world.
    """
    init_object_state()

    msg = DesignatorResolutionStart()
    msg.designator_id = str(hash(node))
    msg.json_designator = designator_to_json(node)
    msg.stamp = get_current_ros_time()

    desig_resolution_start.publish(msg)


def resolution_finished_callback(node: PlanNode):
    """
    Callback function that is called when the resolution of a designator finishes. It publishes a message to Knowrob
    containing the designator that has just finished resolving.
    """
    msg = DesignatorResolutionFinished()
    msg.designator_id = str(hash(node.children[-1]))
    msg.json_designator = designator_to_json(node)
    msg.stamp = get_current_ros_time()
    msg.resolved_from_id = str(hash(node))

    desig_resolution_finished.publish(msg)


# Registers the callback to the Type of Action

Plan.add_on_start_callback(resolution_start_callback, ActionNode)
Plan.add_on_start_callback(execution_start_callback, ResolvedActionNode)

Plan.add_on_end_callback(resolution_finished_callback, ActionNode)
Plan.add_on_end_callback(execution_finished_callback, ResolvedActionNode)
